from typing import Dict, List, Optional
import inspect
import importlib
import can
from contextlib import contextmanager
import logging
from docstring_parser import parse


class LogCaptureHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.records = []

    def emit(self, record):
        self.records.append(record)


@contextmanager
def capture_logs(logger_name: str):
    log_handler = LogCaptureHandler()
    target_logger = logging.getLogger(logger_name)
    original_handlers = target_logger.handlers[:]
    original_level = target_logger.level
    try:
        target_logger.handlers.clear()
        target_logger.addHandler(log_handler)
        target_logger.setLevel(logging.WARNING)
        yield log_handler
    finally:
        target_logger.handlers = original_handlers
        target_logger.setLevel(original_level)


class CANInterfaceManager:
    """
    Dynamically discovers available python-can interfaces. It uses the
    'docstring-parser' library to parse their docstrings, finds their
    configuration parameters, and filters out any interfaces that produce
    warnings or errors during discovery.
    """

    def __init__(self):
        self._interfaces = self._discover_interfaces()

    def _discover_interfaces(self):
        interfaces = {}
        for name, (module_name, class_name) in can.interfaces.BACKENDS.items():
            try:
                parsed_doc_dict = {}
                with capture_logs("can") as log_handler:
                    module = importlib.import_module(module_name)
                    bus_class = getattr(module, class_name)

                    init_doc = inspect.getdoc(bus_class.__init__)
                    raw_doc = init_doc if init_doc else inspect.getdoc(bus_class)

                    if raw_doc:
                        parsed = parse(raw_doc)

                        desc_parts = []
                        if parsed.short_description:
                            desc_parts.append(parsed.short_description)
                        if parsed.long_description:
                            desc_parts.append(parsed.long_description)
                        description = "\n\n".join(desc_parts)

                        params_dict = {
                            param.arg_name: {
                                "type_name": param.type_name,
                                "description": param.description or "",
                            }
                            for param in parsed.params
                        }

                        parsed_doc_dict = {
                            "description": description,
                            "params": params_dict,
                        }
                    else:
                        parsed_doc_dict = {"description": "", "params": {}}

                    if log_handler.records:
                        first_warning = log_handler.records[0].getMessage()
                        print(
                            f"Info: Skipping interface '{name}' due to warning: {first_warning}"
                        )
                        continue

                sig = inspect.signature(bus_class.__init__)
                params = {}
                for param in sig.parameters.values():
                    if param.name in ["self", "args", "kwargs", "receive_own_messages"]:
                        continue
                    param_info = {
                        "default": param.default
                        if param.default is not inspect.Parameter.empty
                        else None,
                        "type": param.annotation
                        if param.annotation is not inspect.Parameter.empty
                        else type(param.default),
                    }
                    params[param.name] = param_info

                interfaces[name] = {
                    "class": bus_class,
                    "params": params,
                    "docstring": parsed_doc_dict,
                }

            except (ImportError, AttributeError, OSError, TypeError) as e:
                print(f"Info: Skipping interface '{name}' due to error on load: {e}")
            except Exception as e:
                print(f"Warning: Could not load or inspect CAN interface '{name}': {e}")

        return dict(sorted(interfaces.items()))

    def get_available_interfaces(self) -> List[str]:
        return list(self._interfaces.keys())

    def get_interface_params(self, name: str) -> Optional[Dict]:
        return self._interfaces.get(name, {}).get("params")

    def get_interface_docstring(self, name: str) -> Optional[Dict]:
        """Returns the parsed docstring for the given interface name."""
        return self._interfaces.get(name, {}).get("docstring")



from __future__ import annotations
import cantools
from collections import deque
from dataclasses import dataclass, field
from typing import List, Optional, Any, Dict
import enum
import uuid

from PySide6.QtCore import QAbstractItemModel, QAbstractTableModel, QModelIndex, Qt

from .data_utils import CANFrame, DBCFile
from .co.canopen_utils import CANopenDecoder

# --- Data Structures ---


@dataclass
class DisplayItem:  # Used for Grouped View
    parent: Optional[DisplayItem]
    data_source: Any
    is_signal: bool = False
    children: List[DisplayItem] = field(default_factory=list)
    children_populated: bool = False
    row_in_parent: int = 0


@dataclass
class SignalInfo:
    """Represents a single decoded signal or field."""

    name: str
    value: Any
    unit: str


@dataclass
class DecodingResult:
    """Represents the complete decoding of a message from a single source."""

    source: str  # e.g., "DBC", "PDO", "CANopen"
    name: str  # e.g., "EngineStatus", "TPDO1_Node5", "NMT"
    signals: List[SignalInfo]


# --- Helper Function for Decoding ---


def get_structured_decodings(
    frame: CANFrame,
    dbc_files: List[DBCFile],
    pdo_databases: List[cantools.db.Database],
    canopen_enabled: bool,
) -> List[DecodingResult]:
    """
    Decodes a CAN frame from all available sources and returns structured results.

    This is the centralized decoding function.
    """
    results: List[DecodingResult] = []

    def _process_database(db: cantools.db.Database, source_name: str) -> None:
        """Helper to decode from a cantools database and append to results."""
        try:
            message = db.get_message_by_frame_id(frame.arbitration_id)
            decoded_signals = db.decode_message(
                frame.arbitration_id, frame.data, decode_choices=False
            )

            signal_infos = [
                SignalInfo(
                    name=s.name,
                    value=decoded_signals.get(s.name, "N/A"),
                    unit=s.unit or "",
                )
                for s in message.signals
            ]

            results.append(
                DecodingResult(
                    source=source_name, name=message.name, signals=signal_infos
                )
            )
        except (KeyError, ValueError):
            pass  # Frame not in this database

    # 1. Process regular DBCs
    for dbc in dbc_files:
        if dbc.connection_id is None or dbc.connection_id == frame.connection_id:
            _process_database(dbc.database, "DBC")

    # 2. Process CANopen PDO databases
    for db in pdo_databases:
        _process_database(db, "PDO")

    # 3. Process generic CANopen
    if canopen_enabled:
        if co_info := CANopenDecoder.decode(frame):
            # The 'CANopen Type' becomes the name, the rest are signals
            canopen_type = co_info.pop("CANopen Type", "Unknown")
            signal_infos = [
                SignalInfo(name=k, value=v, unit="") for k, v in co_info.items()
            ]
            results.append(
                DecodingResult(
                    source="CANopen", name=canopen_type, signals=signal_infos
                )
            )

    return results


# --- Models ---
class TraceViewColumn(enum.IntEnum):
    """Defines the columns for the CANTraceModel."""

    TIMESTAMP = 0
    BUS = 1
    DIRECTION = 2
    ID = 3
    TYPE = 4
    DLC = 5
    DATA = 6
    DECODED = 7


class CANTraceModel(QAbstractTableModel):
    def __init__(self):
        super().__init__()

        # Programmatically create headers from the Enum
        self.headers = [col.name.replace("_", " ").title() for col in TraceViewColumn]
        self.headers[TraceViewColumn.DIRECTION] = "Rx/Tx"  # Custom header name

        self.frames: deque[CANFrame] = deque(maxlen=10000)
        self.dbc_files: List[DBCFile] = []
        self.pdo_databases: List[cantools.db.Database] = []
        self.canopen_enabled = True

    def set_data(self, frames: List[CANFrame]):
        self.beginResetModel()
        self.frames.clear()
        self.frames.extend(frames)
        self.endResetModel()

    def set_config(
        self,
        dbs: List[DBCFile],
        co_enabled: bool,
        pdo_dbs: List[cantools.db.Database] | None = None,
    ):
        self.dbc_files = dbs
        self.canopen_enabled = co_enabled
        self.pdo_databases = pdo_dbs or []
        self.layoutChanged.emit()

    def rowCount(self, p=QModelIndex()):
        return len(self.frames)

    def columnCount(self, p=QModelIndex()):
        return len(TraceViewColumn)

    def headerData(self, s, o, r):
        if o == Qt.Horizontal and r == Qt.DisplayRole:
            return self.headers[s]

    def data(self, index, role):
        if not index.isValid() or role != Qt.DisplayRole:
            return None

        frame = self.frames[index.row()]
        try:
            # Convert column index to our symbolic Enum
            col = TraceViewColumn(index.column())
        except ValueError:
            return None

        if col == TraceViewColumn.TIMESTAMP:
            return f"{frame.timestamp:.6f}"
        if col == TraceViewColumn.BUS:
            return str(frame.bus)
        if col == TraceViewColumn.DIRECTION:
            return "Rx" if frame.is_rx else "Tx"
        if col == TraceViewColumn.ID:
            return f"0x{frame.arbitration_id:X}"
        if col == TraceViewColumn.TYPE:
            return ("Ext" if frame.is_extended else "Std") + (
                " RTR" if frame.is_remote else ""
            )
        if col == TraceViewColumn.DLC:
            return str(frame.dlc)
        if col == TraceViewColumn.DATA:
            return frame.data.hex(" ")
        if col == TraceViewColumn.DECODED:
            return self._decode_frame(frame)

        return None

    def _decode_frame(self, frame: CANFrame) -> str:
        """Formats structured decoding results into a string for the Trace view."""
        structured_results = get_structured_decodings(
            frame, self.dbc_files, self.pdo_databases, self.canopen_enabled
        )

        output_strings = []
        for result in structured_results:
            signals_str = " ".join(f"{s.name}={s.value}" for s in result.signals)
            output_strings.append(f"{result.source}: {result.name} | {signals_str}")

        return " || ".join(output_strings)


class GroupedViewColumn(enum.IntEnum):
    """Defines the columns for the CANGroupedModel."""

    ID = 0
    BUS = 1
    NAME = 2
    DLC = 3
    DATA = 4
    CYCLE_TIME = 5
    COUNT = 6


class CANGroupedModel(QAbstractItemModel):
    def __init__(self):
        super().__init__()
        # Programmatically create headers from the Enum, this replaces underscores with spaces for display
        self.headers = [col.name.replace("_", " ").title() for col in GroupedViewColumn]
        self.top_level_items: List[DisplayItem] = []
        self.dbc_files: List[DBCFile] = []
        self.pdo_databases: List[cantools.db.Database] = []
        self.canopen_enabled = True
        self.frame_counts = {}
        self.timestamps = {}
        self.item_map = {}

    def set_config(
        self,
        dbs: List[DBCFile],
        co_enabled: bool,
        pdo_dbs: List[cantools.db.Database] | None = None,
    ):
        self.dbc_files = dbs
        self.canopen_enabled = co_enabled
        self.pdo_databases = pdo_dbs or []
        self.layoutChanged.emit()

    def columnCount(self, p=QModelIndex()):
        return len(GroupedViewColumn)

    def headerData(self, s, o, r):
        if o == Qt.Horizontal and r == Qt.DisplayRole:
            return self.headers[s]

    def rowCount(self, p=QModelIndex()):
        if not p.isValid():
            return len(self.top_level_items)
        return (
            len(p.internalPointer().children)
            if p.internalPointer().children_populated
            else 0
        )

    def index(self, r, c, p=QModelIndex()):
        if not self.hasIndex(r, c, p):
            return QModelIndex()
        parent = p.internalPointer() if p.isValid() else None
        items = self.top_level_items if not parent else parent.children
        return self.createIndex(r, c, items[r]) if r < len(items) else QModelIndex()

    def parent(self, i):
        if not i.isValid():
            return QModelIndex()
        parent = i.internalPointer().parent
        return (
            self.createIndex(parent.row_in_parent, 0, parent)
            if parent
            else QModelIndex()
        )

    def hasChildren(self, p=QModelIndex()):
        if not p.isValid():
            return True
        item = p.internalPointer()
        if item.is_signal:
            return False
        if item.children_populated:
            return len(item.children) > 0

        # Check for any possible signals from any source
        signals = self._decode_frame_to_signals(item.data_source)
        return len(signals) > 0

    def canFetchMore(self, p: QModelIndex):
        return not p.internalPointer().children_populated if p.isValid() else False

    def fetchMore(self, p: QModelIndex):
        item = p.internalPointer()
        if item.children_populated:
            return

        signals = self._decode_frame_to_signals(item.data_source)
        if not signals:
            item.children_populated = True
            return

        self.beginInsertRows(p, 0, len(signals) - 1)
        item.children = [
            DisplayItem(item, s, True, row_in_parent=i) for i, s in enumerate(signals)
        ]
        item.children_populated = True
        self.endInsertRows()

    def _decode_frame_to_signals(self, frame: CANFrame) -> List[Dict]:
        """Formats structured decoding results into a list of dicts for the Grouped view."""
        structured_results = get_structured_decodings(
            frame, self.dbc_files, self.pdo_databases, self.canopen_enabled
        )

        all_signals = []
        for result in structured_results:
            # Add a header for the decoding source if there are multiple
            if len(structured_results) > 1:
                header_name = f"--- {result.source}: {result.name} ---"
                all_signals.append({"name": header_name, "value": "", "unit": ""})

            # Add the actual signals
            for sig_info in result.signals:
                all_signals.append(
                    {
                        "name": sig_info.name,
                        "value": sig_info.value,
                        "unit": sig_info.unit,
                    }
                )

        return all_signals

    def clear_frames(self):
        self.beginResetModel()
        self.top_level_items.clear()
        self.frame_counts.clear()
        self.timestamps.clear()
        self.item_map.clear()
        self.endResetModel()

    def update_frames(self, frames: List[CANFrame]):
        if not frames:
            return
        self.beginResetModel()
        for frame in frames:
            if not frame.is_rx:  # TODO : make this configurable ?
                continue  # Skip Tx frames

            # Use a unique key for each connection_id/ID pair
            item_key = (frame.connection_id, frame.arbitration_id)

            self.frame_counts[item_key] = self.frame_counts.get(item_key, 0) + 1
            if item_key not in self.timestamps:
                self.timestamps[item_key] = deque(maxlen=10)
            self.timestamps[item_key].append(frame.timestamp)

            if item_key not in self.item_map:
                item = DisplayItem(parent=None, data_source=frame)
                item.row_in_parent = len(self.top_level_items)
                self.top_level_items.append(item)
                self.item_map[item_key] = item
            else:
                item = self.item_map[item_key]
                item.data_source = frame
                if item.children_populated:
                    # Mark for refetching on next expansion
                    item.children.clear()
                    item.children_populated = False
        self.endResetModel()

    def _get_message_name(self, can_id: int, connection_id: uuid.UUID) -> str:
        """Helper to get a message name from any available database."""
        # Prioritize regular DBCs
        for db in self.dbc_files:
            if db.connection_id is None or db.connection_id == connection_id:
                try:
                    return db.database.get_message_by_frame_id(can_id).name
                except KeyError:
                    pass
        # Then PDO DBCs
        for db in self.pdo_databases:
            try:
                return db.get_message_by_frame_id(can_id).name
            except KeyError:
                pass
        return ""

    def data(self, index, role):
        if not index.isValid():
            return None

        item: DisplayItem = index.internalPointer()
        try:
            # Convert column index to our symbolic Enum
            col = GroupedViewColumn(index.column())
        except ValueError:
            return None  # Should not happen with a valid index

        # --- Handle UserRole for Sorting ---
        if role == Qt.UserRole:
            if item.is_signal:
                return None

            item_key = (item.data_source.connection_id, item.data_source.arbitration_id)
            if col == GroupedViewColumn.ID:
                return item.data_source.arbitration_id
            if col == GroupedViewColumn.COUNT:
                return self.frame_counts.get(item_key, 0)

            return None

        # --- Handle DisplayRole for UI Text ---
        if role != Qt.DisplayRole:
            return None

        if item.is_signal:
            sig = item.data_source
            if col == GroupedViewColumn.ID:
                return f"  └ {sig['name']}"
            if col == GroupedViewColumn.NAME:
                return sig.get("unit", "")
            if col == GroupedViewColumn.DATA:
                return f"{sig['value']}"
        else:
            frame: CANFrame = item.data_source
            can_id = frame.arbitration_id
            item_key = (frame.connection_id, frame.arbitration_id)

            if col == GroupedViewColumn.ID:
                return f"0x{can_id:X}"
            if col == GroupedViewColumn.BUS:
                return str(frame.bus)
            if col == GroupedViewColumn.NAME:
                return self._get_message_name(can_id, frame.connection_id)
            if col == GroupedViewColumn.DLC:
                return str(frame.dlc)
            if col == GroupedViewColumn.DATA:
                return frame.data.hex(" ").upper()
            if col == GroupedViewColumn.COUNT:
                return str(self.frame_counts.get(item_key, 0))
            if col == GroupedViewColumn.CYCLE_TIME:
                ts_list = self.timestamps.get(item_key, [])
                if len(ts_list) > 1:
                    cycle_times = [
                        ts_list[i] - ts_list[i - 1] for i in range(1, len(ts_list))
                    ]
                    avg_cycle_ms = sum(cycle_times) / len(cycle_times) * 1000
                    return f"{avg_cycle_ms:.1f}"
                return "-"

        return None



from dataclasses import dataclass, field, asdict
from typing import Dict, List, Any, Optional
from pathlib import Path
import enum
import inspect
import cantools
import uuid

from .interfaces_utils import CANInterfaceManager


@dataclass
class CANFrame:
    timestamp: float
    arbitration_id: int
    data: bytes
    dlc: int
    is_extended: bool = False
    is_error: bool = False
    is_remote: bool = False
    bus: str | None = None
    connection_id: Optional[uuid.UUID] = None
    is_rx: bool = True


@dataclass
class DBCFile:
    path: Path
    database: object
    enabled: bool = True
    connection_id: Optional[uuid.UUID] = None


@dataclass
class CANFrameFilter:
    name: str = "New Filter"
    enabled: bool = True
    connection_id: Optional[uuid.UUID] = None
    min_id: int = 0x000
    max_id: int = 0x7FF
    mask: int = 0x7FF
    accept_extended: bool = True
    accept_standard: bool = True
    accept_data: bool = True
    accept_remote: bool = True

    def matches(self, frame: CANFrame) -> bool:
        # This will need to be updated to use connection_id from frame
        # For now, keeping it as channel for compatibility until frame is updated
        if self.connection_id is not None and self.connection_id != frame.connection_id:
            return False
        if frame.is_extended and not self.accept_extended:
            return False
        if not frame.is_extended and not self.accept_standard:
            return False
        if frame.is_remote and not self.accept_remote:
            return False
        if not frame.is_remote and not self.accept_data:
            return False
        return self.min_id <= (frame.arbitration_id & self.mask) <= self.max_id


@dataclass
class CANopenNode:
    path: Path
    node_id: int
    enabled: bool = True
    connection_id: Optional[uuid.UUID] = None
    pdo_decoding_enabled: bool = True

    def to_dict(self) -> Dict:
        return {
            "path": str(self.path),
            "node_id": self.node_id,
            "enabled": self.enabled,
            "connection_id": str(self.connection_id) if self.connection_id else None,
            "pdo_decoding_enabled": self.pdo_decoding_enabled,
        }

    @classmethod
    def from_dict(cls, data: Dict) -> "CANopenNode":
        path = Path(data["path"])
        if not path.exists():
            raise FileNotFoundError(f"EDS/DCF file not found: {path}")
        return cls(
            path=path,
            node_id=data["node_id"],
            enabled=data["enabled"],
            connection_id=uuid.UUID(data["connection_id"])
            if data.get("connection_id")
            else None,
            pdo_decoding_enabled=data.get("pdo_decoding_enabled", True),
        )


@dataclass
class Connection:
    id: uuid.UUID = field(default_factory=uuid.uuid4)
    name: str = ""
    interface: str = "virtual"
    config: Dict[str, Any] = field(default_factory=dict)
    enabled: bool = True

    def __post_init__(self):
        if not self.name:
            self.name = "vcan0"
        if not self.config:
            if self.interface == "virtual":
                self.config["channel"] = self.name

    def to_dict(self) -> Dict:
        serializable_config = {
            k: v.name if isinstance(v, enum.Enum) else v for k, v in self.config.items()
        }
        return {
            "id": str(self.id),
            "name": self.name,
            "interface": self.interface,
            "config": serializable_config,
            "enabled": self.enabled,
        }

    @classmethod
    def from_dict(
        cls, data: Dict, interface_manager: CANInterfaceManager
    ) -> "Connection":
        interface = data.get("interface", "virtual")
        config_from_file = data.get("config", {})
        hydrated_config = {}
        param_defs = interface_manager.get_interface_params(interface)

        if param_defs:
            for key, value in config_from_file.items():
                if key not in param_defs:
                    hydrated_config[key] = value
                    continue

                param_info = param_defs[key]
                expected_type = param_info.get("type")
                is_enum = False
                try:
                    if inspect.isclass(expected_type) and issubclass(
                        expected_type, enum.Enum
                    ):
                        is_enum = True
                except TypeError:
                    pass

                if is_enum and isinstance(value, str):
                    try:
                        hydrated_config[key] = expected_type[value]
                    except KeyError:
                        hydrated_config[key] = param_info.get("default")
                else:
                    hydrated_config[key] = value
        else:
            hydrated_config = config_from_file

        return cls(
            id=uuid.UUID(data["id"]) if data.get("id") else uuid.uuid4(),
            name=data.get("name", "default"),
            interface=interface,
            config=hydrated_config,
            enabled=data.get("enabled", True),
        )


@dataclass
class Project:
    connections: List[Connection] = field(default_factory=list)
    dbcs: List[DBCFile] = field(default_factory=list)
    filters: List[CANFrameFilter] = field(default_factory=list)
    canopen_enabled: bool = False
    canopen_nodes: List[CANopenNode] = field(default_factory=list)

    def get_active_dbcs(self) -> List[DBCFile]:
        return [dbc for dbc in self.dbcs if dbc.enabled]

    def get_active_filters(self) -> List[CANFrameFilter]:
        return [f for f in self.filters if f.enabled and f.connection_id is not None]

    def get_active_connections(self) -> List[Connection]:
        return [c for c in self.connections if c.enabled]

    def get_connection_name(self, connection_id: uuid.UUID) -> str:
        for conn in self.connections:
            if conn.id == connection_id:
                return conn.name
        return None

    def to_dict(self) -> Dict:
        return {
            "connections": [c.to_dict() for c in self.connections],
            "dbcs": [
                {
                    "path": str(dbc.path),
                    "enabled": dbc.enabled,
                    "connection_id": str(dbc.connection_id)
                    if dbc.connection_id
                    else None,
                }
                for dbc in self.dbcs
            ],
            "filters": [asdict(f) for f in self.filters],
            "canopen_enabled": self.canopen_enabled,
            "canopen_nodes": [node.to_dict() for node in self.canopen_nodes],
        }

    @classmethod
    def from_dict(cls, data: Dict, interface_manager: CANInterfaceManager) -> "Project":
        project = cls()

        if "can_interface" in data:
            # This part handles legacy project files where connections were not
            # explicitly defined but inferred from 'can_interface' and 'can_config'.
            # We create a new Connection object with a UUID for these.
            conn = Connection.from_dict(
                {
                    "name": data["can_interface"],
                    "interface": data["can_interface"],
                    "config": data.get("can_config", {}),
                    "enabled": True,
                    "id": str(uuid.uuid4()),  # Assign a new UUID for legacy connections
                },
                interface_manager,
            )
            project.connections.append(conn)

        for conn_data in data.get("connections", []):
            try:
                project.connections.append(
                    Connection.from_dict(conn_data, interface_manager)
                )
            except Exception as e:
                print(f"Warning: Could not load connection from project: {e}")

        if not project.connections:
            project.connections.append(Connection())

        project.canopen_enabled = data.get("canopen_enabled", False)

        for node_data in data.get("canopen_nodes", []):
            try:
                project.canopen_nodes.append(CANopenNode.from_dict(node_data))
            except Exception as e:
                print(f"Warning: Could not load CANopen node from project: {e}")

        project.filters = [
            CANFrameFilter(**f_data) for f_data in data.get("filters", [])
        ]
        for dbc_data in data.get("dbcs", []):
            try:
                path = Path(dbc_data["path"])
                if not path.exists():
                    raise FileNotFoundError(f"DBC file not found: {path}")
                db = cantools.database.load_file(path)
                project.dbcs.append(
                    DBCFile(
                        path,
                        db,
                        dbc_data.get("enabled", True),
                        uuid.UUID(dbc_data["connection_id"])
                        if dbc_data.get("connection_id")
                        else None,
                    )
                )
            except Exception as e:
                print(f"Warning: Could not load DBC from project file: {e}")
        return project



"""
A standalone SDO client for handling SDO communication.
"""

import asyncio
import struct
import uuid

import can
import canopen


class SdoAbortedError(Exception):
    """Raised when an SDO transfer is aborted by the server."""

    def __init__(self, code):
        self.code = code
        super().__init__(f"SDO Abort Code: 0x{code:08X}")


class SdoClient:
    """A client for handling SDO communication."""

    def __init__(self, node_id: int, connection_id: uuid.UUID, frame_to_send_callback):
        self.node_id = node_id
        self.connection_id = connection_id
        self.frame_to_send_callback = frame_to_send_callback
        self.response_queue = asyncio.Queue(maxsize=127)
        self.transfer_abort_event = asyncio.Event()

    async def on_frame_rx(self, frame: can.Message):
        """Handle received SDO frames by placing them in the response queue."""
        if frame.arbitration_id != (0x580 + self.node_id):
            return

        try:
            self.response_queue.put_nowait(frame)
        except asyncio.QueueFull:
            print("SDO response queue is full, dropping frame.")

    async def read(self, index: int, subindex: int) -> bytes:
        """Read a value from the object dictionary."""
        self.transfer_abort_event.clear()
        while not self.response_queue.empty():
            self.response_queue.get_nowait()

        request = bytearray(8)
        command_specifier = canopen.sdo.constants.REQUEST_UPLOAD
        struct.pack_into("<B H B", request, 0, command_specifier, index, subindex)

        self.frame_to_send_callback(
            can.Message(
                arbitration_id=0x600 + self.node_id,
                is_extended_id=False,
                dlc=8,
                data=request,
            ),
            self.connection_id,
        )

        response_frame = await asyncio.wait_for(self.response_queue.get(), timeout=2.0)
        cs, res_index, res_subindex, payload = struct.unpack_from(
            "<B H B 4s", response_frame.data
        )

        if cs == 0x80:
            abort_code = struct.unpack_from("<L", payload)[0]
            raise SdoAbortedError(abort_code)

        if res_index != index or res_subindex != subindex:
            raise ValueError("Received SDO response for a different object.")

        if cs & 0x02:  # Expedited transfer
            size_indicated = (cs >> 2) & 0x03
            num_bytes = 4 - size_indicated
            return payload[:num_bytes]

        if cs & 0x01:  # Segmented transfer initiated
            total_size = struct.unpack_from("<L", payload)[0]
            received_data = bytearray()
            toggle_bit = 0

            while len(received_data) < total_size:
                if self.transfer_abort_event.is_set():
                    raise asyncio.CancelledError("Transfer aborted by user")

                segment_req_cs = 0x60 | (toggle_bit << 4)
                self.frame_to_send_callback(
                    can.Message(
                        arbitration_id=0x600 + self.node_id,
                        is_extended_id=False,
                        dlc=8,
                        data=bytes([segment_req_cs, 0, 0, 0, 0, 0, 0, 0]),
                    ),
                    self.connection_id,
                )

                segment_frame = await asyncio.wait_for(
                    self.response_queue.get(), timeout=2.0
                )
                seg_cs = segment_frame.data[0]
                seg_payload = segment_frame.data[1:]

                if seg_cs == 0x80:
                    abort_code = struct.unpack_from("<L", segment_frame.data[4:])[0]
                    raise SdoAbortedError(abort_code)

                if (seg_cs >> 4) & 1 != toggle_bit:
                    raise ValueError("SDO protocol error: toggle bit mismatch.")

                bytes_in_segment = 7 - ((seg_cs >> 1) & 0x7)
                received_data.extend(seg_payload[:bytes_in_segment])

                if seg_cs & 0x01:
                    break

                toggle_bit ^= 1
            return received_data

        raise ValueError(f"Invalid SDO initiation response (CS=0x{cs:02X})")

    async def write(self, index: int, subindex: int, data: bytes):
        """Write a value to the object dictionary."""
        self.transfer_abort_event.clear()
        while not self.response_queue.empty():
            self.response_queue.get_nowait()

        total_size = len(data)
        init_request = bytearray(8)

        if total_size <= 4:
            cs = 0x23 | ((4 - total_size) << 2)
            payload = bytearray(data)
            payload.extend(b"\x00" * (4 - total_size))
            struct.pack_into("<B H B 4s", init_request, 0, cs, index, subindex, payload)
        else:
            cs = 0x21
            struct.pack_into(
                "<B H B L", init_request, 0, cs, index, subindex, total_size
            )

        self.frame_to_send_callback(
            can.Message(
                arbitration_id=0x600 + self.node_id,
                is_extended_id=False,
                dlc=8,
                data=init_request,
            ),
            self.connection_id,
        )

        response_frame = await asyncio.wait_for(self.response_queue.get(), timeout=2.0)
        cs, res_index, res_subindex, _ = struct.unpack_from(
            "<B H B 4s", response_frame.data
        )

        if cs == 0x80:
            abort_code = struct.unpack_from("<L", response_frame.data[4:])[0]
            raise SdoAbortedError(abort_code)

        if cs != 0x60:
            raise ValueError(f"Invalid SDO initiation response (CS=0x{cs:02X})")

        if total_size > 4:
            toggle_bit = 0
            bytes_sent = 0
            while bytes_sent < total_size:
                if self.transfer_abort_event.is_set():
                    raise asyncio.CancelledError("Transfer aborted by user")

                chunk = data[bytes_sent : bytes_sent + 7]
                bytes_in_segment = len(chunk)
                bytes_sent += bytes_in_segment

                is_last_segment = 1 if bytes_sent >= total_size else 0
                num_unused_bytes = 7 - bytes_in_segment

                seg_cs = (toggle_bit << 4) | (num_unused_bytes << 1) | is_last_segment

                seg_payload = bytearray(chunk)
                seg_payload.extend(b"\x00" * num_unused_bytes)

                seg_request = bytearray([seg_cs]) + seg_payload

                self.frame_to_send_callback(
                    can.Message(
                        arbitration_id=0x600 + self.node_id,
                        dlc=8,
                        data=seg_request,
                    ),
                    self.connection_id,
                )

                seg_resp_frame = await asyncio.wait_for(
                    self.response_queue.get(), timeout=2.0
                )
                ack_cs = seg_resp_frame.data[0]

                if ack_cs == 0x80:
                    abort_code = struct.unpack_from("<L", seg_resp_frame.data[4:])[0]
                    raise SdoAbortedError(abort_code)

                if (ack_cs & 0x10) != (toggle_bit << 4):
                    raise ValueError("SDO protocol error: toggle bit mismatch.")

                toggle_bit ^= 1



import asyncio
import pytest
import uuid
import struct
from unittest.mock import MagicMock

import can
from canpeek.co.sdo_client import SdoClient, SdoAbortedError

# Mark all tests in this file as asyncio
pytestmark = pytest.mark.asyncio

NODE_ID = 0x42
CONNECTION_ID = uuid.uuid4()


@pytest.fixture
def mock_send_callback():
    """Fixture for the mock send callback."""
    return MagicMock()


@pytest.fixture
def sdo_client(mock_send_callback):
    """Fixture for the SdoClient."""
    return SdoClient(NODE_ID, CONNECTION_ID, mock_send_callback)


async def queue_response(client: SdoClient, data: bytes, delay: float = 0):
    """Helper to queue a response frame on the client's queue."""
    await asyncio.sleep(delay)
    await client.on_frame_rx(can.Message(arbitration_id=0x580 + NODE_ID, data=data))


# --- Read Tests ---


async def test_read_expedited(sdo_client: SdoClient, mock_send_callback: MagicMock):
    """Test a successful expedited SDO read."""
    index, subindex = 0x1008, 0x00
    test_data = b"test"
    response_payload = bytearray(test_data)
    response_payload.extend(b"\x00" * (4 - len(test_data)))

    # CS = 0x43 (4 bytes, expedited, size indicated)
    response_cs = 0x43
    response = bytearray(
        struct.pack("<B H B 4s", response_cs, index, subindex, response_payload)
    )

    # Create a task for the read operation and a task to queue the response
    read_task = asyncio.create_task(sdo_client.read(index, subindex))
    asyncio.create_task(queue_response(sdo_client, response))

    # Wait for the read operation to complete
    result = await read_task

    assert result == test_data
    mock_send_callback.assert_called_once()
    sent_msg = mock_send_callback.call_args[0][0]
    assert sent_msg.arbitration_id == 0x600 + NODE_ID
    assert sent_msg.data[0] == 0x40  # REQUEST_UPLOAD


async def test_read_segmented(sdo_client: SdoClient, mock_send_callback: MagicMock):
    """Test a successful segmented SDO read."""
    index, subindex = 0x1009, 0x00
    long_data = b"this is a long data string for testing"

    # --- Server Responses ---
    # 1. Initiation response
    init_response_cs = 0x41  # Segmented, size indicated
    init_response = bytearray(
        struct.pack("<B H B L", init_response_cs, index, subindex, len(long_data))
    )

    # 2. Segment responses
    segment_responses = []
    toggle = 0
    bytes_sent = 0
    while bytes_sent < len(long_data):
        chunk = long_data[bytes_sent : bytes_sent + 7]
        bytes_in_segment = len(chunk)
        bytes_sent += bytes_in_segment
        is_last = 1 if bytes_sent >= len(long_data) else 0

        seg_cs = (toggle << 4) | ((7 - bytes_in_segment) << 1) | is_last
        payload = bytearray(chunk)
        payload.extend(b"\x00" * (7 - len(chunk)))
        segment_responses.append(bytearray([seg_cs]) + payload)
        toggle ^= 1

    # --- Test Execution ---
    read_task = asyncio.create_task(sdo_client.read(index, subindex))

    # Queue the responses
    await queue_response(sdo_client, init_response, delay=0.01)
    for resp in segment_responses:
        await queue_response(sdo_client, resp, delay=0.01)

    result = await read_task

    assert result == long_data
    # Check that the correct number of requests were sent (1 initiate + num_segments)
    assert mock_send_callback.call_count == 1 + len(segment_responses)


# --- Write Tests ---


async def test_write_expedited(sdo_client: SdoClient, mock_send_callback: MagicMock):
    """Test a successful expedited SDO write."""
    index, subindex = 0x2000, 0x01
    data_to_write = b"OK"

    # Server response: success
    response_cs = 0x60  # download success
    response = bytearray(struct.pack("<B H B L", response_cs, index, subindex, 0))

    write_task = asyncio.create_task(sdo_client.write(index, subindex, data_to_write))
    await queue_response(sdo_client, response, delay=0.01)

    await write_task  # Should complete without error

    mock_send_callback.assert_called_once()
    sent_msg = mock_send_callback.call_args[0][0]
    assert sent_msg.arbitration_id == 0x600 + NODE_ID
    # Check for expedited download request CS
    assert sent_msg.data[0] & 0xF3 == 0x23


# --- Error Handling Tests ---


async def test_sdo_abort(sdo_client: SdoClient):
    """Test that SdoAbortedError is raised on abort."""
    index, subindex = 0x1010, 0x00
    abort_code = 0x06090011  # Object does not exist

    response = bytearray(struct.pack("<B H B L", 0x80, index, subindex, abort_code))

    read_task = asyncio.create_task(sdo_client.read(index, subindex))
    await queue_response(sdo_client, response, delay=0.01)

    with pytest.raises(SdoAbortedError) as excinfo:
        await read_task
    assert excinfo.value.code == abort_code


async def test_sdo_timeout(sdo_client: SdoClient):
    """Test that asyncio.TimeoutError is raised on timeout."""
    with pytest.raises(asyncio.TimeoutError):
        await sdo_client.read(0x1018, 0x01)



#!/usr/bin/env python3
"""
CAN Bus Observer GUI
Features:
- Project-based configuration with Tree View
- Highly performant, batched-update Trace/Grouped views
- Multi-DBC and Multi-Filter support, enhanced CANopen decoding
- DBC content viewer
- DBC decoding and signal-based transmitting
- CAN log file saving/loading
- Real-time monitoring
"""

import os
import qt_themes
import sys
import json
from typing import Dict, List, Optional, Any, TYPE_CHECKING
from pathlib import Path
from functools import partial
import inspect
import enum
from . import rc_icons
import asyncio
from qasync import QEventLoop, QApplication
import uuid


__all__ = [
    "rc_icons",  # remove ruff "Remove unused import: `.rc_icons`"
]


from PySide6.QtWidgets import (
    QMainWindow,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QTableWidget,
    QTableWidgetItem,
    QPushButton,
    QLabel,
    QLineEdit,
    QCheckBox,
    QComboBox,
    QSpinBox,
    QGroupBox,
    QFormLayout,
    QHeaderView,
    QFileDialog,
    QMessageBox,
    QMenu,
    QTreeView,
    QTreeWidget,
    QTreeWidgetItem,
    QTableView,
    QToolBar,
    QStyle,
    QDialog,
    QTextEdit,
    QSizePolicy,
    QWidgetAction,
    QInputDialog,
)

from PySide6.QtCore import (
    QTimer,
    Signal,
    Qt,
    QSortFilterProxyModel,
    QSettings,
)
from PySide6.QtGui import (
    QAction,
    QKeyEvent,
    QIcon,
    QPixmap,
    QColor,
    QActionGroup,
    QFont,
    QFontDatabase,
)

import can
import cantools

import PySide6QtAds as QtAds

from .co.canopen_utils import (
    CANopenNode,
    CANopenNodeEditor,
    CANopenRootEditor,
    PDOEditor,
    PDODatabaseManager,
    ObjectDictionaryViewer,
)
from .co.nmt_editor import NMTSender

from .data_utils import (
    CANFrame,
    Project,
    CANFrameFilter,
    DBCFile,
    CANInterfaceManager,
    Connection,
)

from .can_utils import CANAsyncReader

from .models import CANTraceModel, CANGroupedModel, GroupedViewColumn

if TYPE_CHECKING:
    from __main__ import ProjectExplorer, CANBusObserver

# Workaround for qt-ads on wayland https://github.com/githubuser0xFFFF/Qt-Advanced-Docking-System/issues/714
if os.environ.get("XDG_SESSION_TYPE") == "wayland":
    print("Workaround for qt-ads on Wayland")
    os.environ["QT_QPA_PLATFORM"] = "xcb"

# --- Data Structures ---
TRACE_BUFFER_LIMIT = 5000


# --- UI Classes ---
class DBCEditor(QWidget):
    message_to_transmit = Signal(object)
    project_changed = Signal()

    def __init__(self, dbc_file: DBCFile, project: Project):
        super().__init__()
        self.dbc_file = dbc_file
        self.project = project
        self.sorted_messages = []  # Store sorted messages for transmission
        self.setup_ui()

    def setup_ui(self):
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        group = QGroupBox(f"DBC Content: {self.dbc_file.path.name}")

        form_layout = QFormLayout()

        self.channel_combo = QComboBox()
        self.channel_combo.addItem("All", None)  # "All" option with None as data
        self.connection_map = {conn.name: conn.id for conn in self.project.connections}
        for conn in self.project.connections:
            self.channel_combo.addItem(conn.name, conn.id)

        if self.dbc_file.connection_id:
            if self.dbc_file.connection_id == -1:
                self.channel_combo.setCurrentIndex(-1)
            else:
                # Find the name corresponding to the stored connection_id
                for name, conn_id in self.connection_map.items():
                    if conn_id == self.dbc_file.connection_id:
                        self.channel_combo.setCurrentText(name)
                        break
        else:
            self.channel_combo.setCurrentIndex(0)  # Select "All"

        self.channel_combo.currentTextChanged.connect(self._on_channel_changed)
        form_layout.addRow("Channel:", self.channel_combo)

        layout = QVBoxLayout(group)
        layout.addLayout(form_layout)
        main_layout.addWidget(group)

        self.table = QTableWidget()
        self.table.setEditTriggers(QTableWidget.NoEditTriggers)
        self.table.setColumnCount(4)
        self.table.setHorizontalHeaderLabels(["Message", "ID (hex)", "DLC", "Signals"])
        layout.addWidget(self.table)

        self.table.setContextMenuPolicy(Qt.CustomContextMenu)
        self.table.customContextMenuRequested.connect(self.open_context_menu)

        self.populate_table()
        self.table.resizeColumnsToContents()

    def _on_channel_changed(self, text: str):
        selected_id = self.channel_combo.currentData()
        self.dbc_file.connection_id = selected_id
        self.project_changed.emit()

    def populate_table(self):
        # Store the sorted list in the instance variable
        self.sorted_messages = sorted(
            self.dbc_file.database.messages, key=lambda m: m.frame_id
        )
        self.table.setRowCount(len(self.sorted_messages))
        for r, m in enumerate(self.sorted_messages):
            self.table.setItem(r, 0, QTableWidgetItem(m.name))
            self.table.setItem(r, 1, QTableWidgetItem(f"0x{m.frame_id:X}"))
            self.table.setItem(r, 2, QTableWidgetItem(str(m.length)))
            self.table.setItem(
                r, 3, QTableWidgetItem(", ".join(s.name for s in m.signals))
            )

    # --- Add these two new methods ---
    def open_context_menu(self, position):
        """Creates and shows the context menu."""
        item = self.table.itemAt(position)
        if not item:
            return

        row = item.row()
        message = self.sorted_messages[row]

        menu = QMenu()
        action = QAction(f"Add '{message.name}' to Transmit Panel", self)
        action.triggered.connect(lambda: self._emit_transmit_signal(row))
        menu.addAction(action)

        menu.exec(self.table.viewport().mapToGlobal(position))

    def _emit_transmit_signal(self, row: int):
        """Emits the signal with the selected message object."""
        message = self.sorted_messages[row]
        self.message_to_transmit.emit(message)


class FilterEditor(QWidget):
    filter_changed = Signal()

    def __init__(self, can_filter: CANFrameFilter, project: Project):
        super().__init__()
        self.filter = can_filter
        self.project = project
        self.setup_ui()

    def setup_ui(self):
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        group = QGroupBox("Filter Properties")
        layout = QFormLayout(group)
        main_layout.addWidget(group)
        self.name_edit = QLineEdit(self.filter.name)
        layout.addRow("Name:", self.name_edit)

        self.channel_combo = QComboBox()
        self.channel_combo.addItem("All", None)  # "All" option with None as data
        self.connection_map = {conn.name: conn.id for conn in self.project.connections}
        for conn in self.project.connections:
            self.channel_combo.addItem(conn.name, conn.id)

        if self.filter.connection_id:
            if self.filter.connection_id == -1:
                self.channel_combo.setCurrentIndex(-1)
            else:
                # Find the name corresponding to the stored connection_id
                for name, conn_id in self.connection_map.items():
                    if conn_id == self.filter.connection_id:
                        self.channel_combo.setCurrentText(name)
                        break
        else:
            self.channel_combo.setCurrentIndex(0)  # Select "All"

        self.channel_combo.currentTextChanged.connect(self._update_filter)
        layout.addRow("Channel:", self.channel_combo)

        id_layout = QHBoxLayout()
        self.min_id_edit = QLineEdit(f"0x{self.filter.min_id:X}")
        self.max_id_edit = QLineEdit(f"0x{self.filter.max_id:X}")
        self.mask_edit = QLineEdit(f"0x{self.filter.mask:X}")
        id_layout.addWidget(QLabel("Min:"))
        id_layout.addWidget(self.min_id_edit)
        id_layout.addWidget(QLabel("Max:"))
        id_layout.addWidget(self.max_id_edit)
        id_layout.addWidget(QLabel("Mask:"))
        id_layout.addWidget(self.mask_edit)
        layout.addRow("ID (hex):", id_layout)
        self.standard_cb = QCheckBox("Standard")
        self.standard_cb.setChecked(self.filter.accept_standard)
        self.extended_cb = QCheckBox("Extended")
        self.extended_cb.setChecked(self.filter.accept_extended)
        self.data_cb = QCheckBox("Data")
        self.data_cb.setChecked(self.filter.accept_data)
        self.remote_cb = QCheckBox("Remote")
        self.remote_cb.setChecked(self.filter.accept_remote)
        type_layout = QHBoxLayout()
        type_layout.addWidget(self.standard_cb)
        type_layout.addWidget(self.extended_cb)
        type_layout.addWidget(self.data_cb)
        type_layout.addWidget(self.remote_cb)
        type_layout.addStretch()
        layout.addRow("Frame Types:", type_layout)
        self.name_edit.editingFinished.connect(self._update_filter)
        [
            w.editingFinished.connect(self._update_filter)
            for w in [self.min_id_edit, self.max_id_edit, self.mask_edit]
        ]
        [
            cb.toggled.connect(self._update_filter)
            for cb in [self.standard_cb, self.extended_cb, self.data_cb, self.remote_cb]
        ]

    def _update_filter(self):
        self.filter.name = self.name_edit.text()
        selected_id = self.channel_combo.currentData()
        self.filter.connection_id = selected_id
        try:
            self.filter.min_id = int(self.min_id_edit.text(), 16)
        except ValueError:
            self.min_id_edit.setText(f"0x{self.filter.min_id:X}")
        try:
            self.filter.max_id = int(self.max_id_edit.text(), 16)
        except ValueError:
            self.max_id_edit.setText(f"0x{self.filter.max_id:X}")
        try:
            self.filter.mask = int(self.mask_edit.text(), 16)
        except ValueError:
            self.mask_edit.setText(f"0x{self.filter.mask:X}")
        self.filter.accept_standard = self.standard_cb.isChecked()
        self.filter.accept_extended = self.extended_cb.isChecked()
        self.filter.accept_data = self.data_cb.isChecked()
        self.filter.accept_remote = self.remote_cb.isChecked()
        self.filter_changed.emit()


class DocumentationWindow(QDialog):
    """A separate, non-blocking window for displaying parsed documentation."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Interface Documentation")
        self.setMinimumSize(600, 450)

        layout = QVBoxLayout(self)
        self.text_edit = QTextEdit()
        self.text_edit.setReadOnly(True)
        self.text_edit.setObjectName("documentationViewer")
        layout.addWidget(self.text_edit)

    def set_content(self, interface_name: str, parsed_doc: Dict):
        """
        Updates the window title and content by building an HTML string
        from the parsed docstring dictionary, including type information.
        """
        self.setWindowTitle(f"Documentation for '{interface_name}'")

        html = """
        <style>
            body { font-family: sans-serif; font-size: 14px; }
            p { margin-bottom: 12px; }
            dl { margin-left: 10px; }
            dt { font-weight: bold; color: #af5aed; margin-top: 8px; }
            dt .param-type { font-style: italic; color: #555555; font-weight: normal; }
            dd { margin-left: 20px; margin-bottom: 8px; }
            hr { border: 1px solid #cccccc; }
        </style>
        """

        if parsed_doc and parsed_doc.get("description"):
            desc = parsed_doc["description"].replace("<", "<").replace(">", ">")
            html += f"<p>{desc.replace(chr(10), '<br>')}</p>"

        if parsed_doc and parsed_doc.get("params"):
            html += "<hr><h3>Parameters:</h3>"
            html += "<dl>"
            for name, param_info in parsed_doc["params"].items():
                type_name = param_info.get("type_name")
                description = (
                    param_info.get("description", "")
                    .replace("<", "<")
                    .replace(">", ">")
                )

                # Build the header line (dt) with optional type info
                header = f"<strong>{name}</strong>"
                if type_name:
                    header += f' <span class="param-type">({type_name})</span>'

                html += f"<dt>{header}:</dt><dd>{description}</dd>"
            html += "</dl>"

        if not (
            parsed_doc and (parsed_doc.get("description") or parsed_doc.get("params"))
        ):
            html += "<p>No documentation available.</p>"

        self.text_edit.setHtml(html)


# Fully dynamic editor for connection settings


class ConnectionEditor(QWidget):
    project_changed = Signal()

    def __init__(self, connection: Connection, interface_manager: CANInterfaceManager):
        super().__init__()
        self.connection = connection
        self.interface_manager = interface_manager
        self.dynamic_widgets = {}
        self.docs_window = DocumentationWindow(self)
        self.setup_ui()
        self.interface_combo.setCurrentText(self.connection.interface)
        self._rebuild_dynamic_fields(self.connection.interface)

    def setup_ui(self):
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        group = QGroupBox("Connection Properties")
        self.form_layout = QFormLayout(group)
        main_layout.addWidget(group)

        self.name_edit = QLineEdit(self.connection.name)
        self.form_layout.addRow("Name:", self.name_edit)

        self.interface_combo = QComboBox()
        self.interface_combo.addItems(self.interface_manager.get_available_interfaces())
        self.form_layout.addRow("Interface:", self.interface_combo)

        self.show_docs_button = QPushButton("Show python-can Documentation...")
        self.form_layout.addRow(self.show_docs_button)

        self.dynamic_fields_container = QWidget()
        self.dynamic_layout = QFormLayout(self.dynamic_fields_container)
        self.dynamic_layout.setContentsMargins(0, 0, 0, 0)
        self.form_layout.addRow(self.dynamic_fields_container)

        self.name_edit.editingFinished.connect(self._on_name_changed)
        self.show_docs_button.clicked.connect(self._show_documentation_window)
        self.interface_combo.currentTextChanged.connect(self._on_interface_changed)

    def set_connected_state(self, connected):
        """Enable/disable interface field and settings based on connection state"""
        # Disable interface selection and name editing when connected
        self.interface_combo.setEnabled(not connected)
        # self.name_edit.setEnabled(not connected)

        # Disable all dynamic interface settings when connected
        for widget in self.dynamic_widgets.values():
            if hasattr(widget, "setEnabled"):
                widget.setEnabled(not connected)

    def _on_name_changed(self):
        self.connection.name = self.name_edit.text()
        self.project_changed.emit()

    def _show_documentation_window(self):
        interface_name = self.interface_combo.currentText()
        docstring = self.interface_manager.get_interface_docstring(interface_name)
        self.docs_window.set_content(interface_name, docstring)
        self.docs_window.show()
        self.docs_window.raise_()
        self.docs_window.activateWindow()

    def _on_interface_changed(self, interface_name: str):
        self._rebuild_dynamic_fields(interface_name)
        self.connection.interface = interface_name
        self.project_changed.emit()

    def _rebuild_dynamic_fields(self, interface_name: str):
        parsed_doc = self.interface_manager.get_interface_docstring(interface_name)
        param_docs = parsed_doc.get("params", {}) if parsed_doc else {}
        has_docs = bool(
            parsed_doc and (parsed_doc.get("description") or parsed_doc.get("params"))
        )
        self.show_docs_button.setVisible(has_docs)

        while self.dynamic_layout.count():
            item = self.dynamic_layout.takeAt(0)
            if item.widget():
                item.widget().deleteLater()
        self.dynamic_widgets.clear()

        params = self.interface_manager.get_interface_params(interface_name)
        if not params:
            self._update_project_config()
            return

        for name, info in params.items():
            # Use the default prameter value from python-can if the interface type changed
            if self.connection.interface != interface_name:
                current_value = info.get("default", self.connection.config.get(name))
            else:
                current_value = self.connection.config.get(name, info.get("default"))

            expected_type = info["type"]
            widget = None
            is_enum = False
            try:
                if inspect.isclass(expected_type) and issubclass(
                    expected_type, enum.Enum
                ):
                    is_enum = True
            except TypeError:
                pass

            if is_enum:
                widget = QComboBox()
                widget.setProperty("enum_class", expected_type)
                widget.addItems([m.name for m in list(expected_type)])
                if isinstance(current_value, enum.Enum):
                    widget.setCurrentText(current_value.name)
                elif isinstance(current_value, str) and current_value in [
                    m.name for m in expected_type
                ]:
                    widget.setCurrentText(current_value)
                widget.currentTextChanged.connect(self._update_project_config)
            elif expected_type is bool:
                widget = QCheckBox()
                widget.setChecked(
                    bool(current_value) if current_value is not None else False
                )
                widget.toggled.connect(self._update_project_config)
            elif name == "bitrate" and expected_type is int:
                widget = QSpinBox()
                widget.setRange(1000, 4000000)
                widget.setSuffix(" bps")
                widget.setValue(
                    int(current_value) if current_value is not None else 125000
                )
                widget.valueChanged.connect(self._update_project_config)
            else:
                widget = QLineEdit()
                widget.setText(str(current_value) if current_value is not None else "")
                widget.editingFinished.connect(self._update_project_config)

            if widget:
                tooltip_info = param_docs.get(name)
                if tooltip_info and tooltip_info.get("description"):
                    tooltip_parts = []
                    type_name = tooltip_info.get("type_name")
                    if type_name:
                        tooltip_parts.append(f"({type_name})")
                    tooltip_parts.append(tooltip_info["description"])
                    tooltip_text = " ".join(tooltip_parts)
                    widget.setToolTip(tooltip_text)

                label_text = f"{name.replace('_', ' ').title()}:"
                self.dynamic_layout.addRow(label_text, widget)
                self.dynamic_widgets[name] = widget

        self._update_project_config()

    def _convert_line_edit_text(self, text: str, param_info: Dict) -> Any:
        text = text.strip()
        expected_type = param_info.get("type")
        if text == "" or text.lower() == "none":
            return None

        try:
            if expected_type is int:
                return int(text) if not text.startswith("0x") else int(text, 16)
            if expected_type is float:
                return float(text)
            if expected_type is bool:
                return text.lower() in ("true", "1", "t", "yes", "y")
        except (ValueError, TypeError):
            return None

        return text

    def _update_project_config(self):
        self.connection.interface = self.interface_combo.currentText()
        params = (
            self.interface_manager.get_interface_params(self.connection.interface) or {}
        )
        new_config = {}

        for name, widget in self.dynamic_widgets.items():
            param_info = params.get(name)
            if not param_info:
                continue

            value = None
            try:
                if isinstance(widget, QCheckBox):
                    value = widget.isChecked()
                elif isinstance(widget, QSpinBox):
                    value = widget.value()
                elif isinstance(widget, QComboBox):
                    enum_class = widget.property("enum_class")
                    if enum_class and widget.currentText():
                        value = enum_class[widget.currentText()]
                elif isinstance(widget, QLineEdit):
                    value = self._convert_line_edit_text(widget.text(), param_info)
            except (ValueError, TypeError, KeyError) as e:
                print(f"Warning: Invalid input for '{name}'. Error: {e}")
                value = self.connection.config.get(name)

            if value is not None:
                new_config[name] = value

        self.connection.config.clear()
        self.connection.config.update(new_config)
        self.project_changed.emit()


class PropertiesPanel(QWidget):
    message_to_transmit = Signal(object)

    def __init__(
        self,
        project: Project,
        explorer: "ProjectExplorer",
        interface_manager: CANInterfaceManager,
        main_window: "CANBusObserver",
    ):
        super().__init__()
        self.project = project
        self.explorer = explorer
        self.interface_manager = interface_manager
        self.main_window = main_window
        self.current_widget = None
        self.layout = QVBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.placeholder = QLabel("Select an item to see its properties.")
        self.placeholder.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(self.placeholder)

    def show_properties(self, item: QTreeWidgetItem):
        self.clear()
        data = item.data(0, Qt.UserRole) if item else None
        if isinstance(data, Connection):
            editor = ConnectionEditor(data, self.interface_manager)
            editor.project_changed.connect(self.explorer.rebuild_tree)
            # Set the initial connected state when the editor is shown
            is_connected = data.name in self.main_window.can_readers
            editor.set_connected_state(is_connected)
            self.current_widget = editor
        elif data == "canopen_root":
            editor = CANopenRootEditor(self.project, self.main_window.canopen_network)
            editor.settings_changed.connect(self.explorer.rebuild_tree)
            self.current_widget = editor
        elif isinstance(data, CANopenNode):
            editor = CANopenNodeEditor(data, self.main_window.pdo_manager, self.project)
            editor.node_changed.connect(self.explorer.rebuild_tree)
            editor.node_changed.connect(self.explorer.project_changed.emit)
            self.current_widget = editor
        elif isinstance(data, CANFrameFilter):
            editor = FilterEditor(data, self.project)
            # editor.filter_changed.connect(lambda: item.setText(0, data.name))
            editor.filter_changed.connect(self.explorer.project_changed.emit)
            editor.filter_changed.connect(self.explorer.rebuild_tree)
            self.current_widget = editor
        elif isinstance(data, DBCFile):
            editor = DBCEditor(data, self.project)
            # Connect the editor's signal to the panel's signal
            editor.message_to_transmit.connect(self.message_to_transmit.emit)
            editor.project_changed.connect(self.explorer.rebuild_tree)
            self.current_widget = editor
        elif isinstance(data, tuple) and len(data) == 2 and data[0] == "pdo_content":
            # PDO content viewer for CANopen node
            node = data[1]
            self.current_widget = PDOEditor(node, self.main_window.pdo_manager)
        else:
            self.layout.addWidget(self.placeholder)
            self.placeholder.show()
            return
        self.layout.addWidget(self.current_widget)

    def clear(self):
        if self.current_widget:
            self.current_widget.deleteLater()
            self.current_widget = None
        self.placeholder.hide()


class ProjectExplorer(QWidget):
    project_changed = Signal()

    def __init__(self, project: Project, main_window: "CANBusObserver"):
        super().__init__()
        self.project = project
        self.main_window = main_window
        self.setup_ui()

    def expand_all_items(self):
        """Expands all items in the project tree."""
        self.tree.expandAll()

    def setup_ui(self):
        layout = QVBoxLayout(self)
        self.tree = QTreeWidget()
        self.tree.setHeaderHidden(True)
        layout.addWidget(self.tree)
        self.tree.setContextMenuPolicy(Qt.CustomContextMenu)
        self.tree.customContextMenuRequested.connect(self.open_context_menu)
        self.tree.itemChanged.connect(self.on_item_changed)
        self.rebuild_tree()

    def set_project(self, project: Project):
        self.project = project
        self.rebuild_tree()

    def rebuild_tree(self):
        self.tree.blockSignals(True)
        expanded_items_data = set()
        for i in range(self.tree.topLevelItemCount()):
            root_item = self.tree.topLevelItem(i)
            if root_item.isExpanded():
                expanded_items_data.add(root_item.data(0, Qt.UserRole))
        self.tree.clear()

        self.conn_root = self.add_item(None, "Connections", "connections_root")
        for conn in self.project.connections:
            self.add_item(self.conn_root, conn.name, conn, conn.enabled)

        self.dbc_root = self.add_item(None, "Symbol Files (.dbc)", "dbc_root")
        for dbc in self.project.dbcs:
            conn_name = (
                self.project.get_connection_name(dbc.connection_id)
                if dbc.connection_id
                else "Unassigned"
            )
            self.add_item(
                self.dbc_root,
                dbc.path.name,
                dbc,
                dbc.enabled,
                invalid=(dbc.connection_id == -1),
                tooltip=f"Assigned to: {conn_name}"
                if dbc.connection_id
                else "Unassigned",
            )

        self.filter_root = self.add_item(None, "Message Filters", "filter_root")
        for f in self.project.filters:
            conn_name = (
                self.project.get_connection_name(f.connection_id)
                if f.connection_id
                else "Unassigned"
            )
            self.add_item(
                self.filter_root,
                f.name,
                f,
                f.enabled,
                invalid=(f.connection_id == -1),
                tooltip=f"Assigned to: {conn_name}"
                if f.connection_id
                else "Unassigned",
            )

        self.co_root = self.add_item(None, "CANopen", "canopen_root")
        bus_items = {}
        for node in self.project.canopen_nodes:
            conn_name = (
                self.project.get_connection_name(node.connection_id)
                if node.connection_id
                else None
            )

            if node.connection_id not in bus_items:
                bus_items[node.connection_id] = self.add_item(
                    self.co_root,
                    conn_name if conn_name else "Unassigned",
                    f"canopen_bus_{node.connection_id}",
                    invalid=(conn_name is None),
                )
            self.add_item(
                bus_items[node.connection_id],
                f"{node.path.name} [ID: {node.node_id}]",
                node,
                node.enabled,
            )

        self.tree.expandAll()
        self.tree.blockSignals(False)
        self.project_changed.emit()

    def add_item(
        self, parent, text, data=None, checked=None, invalid=False, tooltip=None
    ):
        item = QTreeWidgetItem(parent or self.tree, [text])
        style = self.style()
        icon = None

        if data == "connections_root":
            icon = style.standardIcon(QStyle.SP_DriveNetIcon)
        elif data == "dbc_root" or data == "filter_root":
            icon = style.standardIcon(QStyle.SP_DirIcon)
        elif data == "canopen_root":
            icon = style.standardIcon(QStyle.SP_ComputerIcon)

        if icon:
            item.setIcon(0, icon)

        if data:
            item.setData(0, Qt.UserRole, data)
        if checked is not None:
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setCheckState(0, Qt.Checked if checked else Qt.Unchecked)

        if invalid:
            item.setBackground(0, QColor(255, 0, 0, 128))
            item.setToolTip(0, tooltip or "This item is invalid and cannot be used.")

        return item

    def on_item_changed(self, item, column):
        if data := item.data(0, Qt.UserRole):
            if isinstance(data, (DBCFile, CANFrameFilter, CANopenNode, Connection)):
                data.enabled = item.checkState(0) == Qt.Checked
            self.project_changed.emit()

    def open_context_menu(self, position):
        menu = QMenu()
        item = self.tree.itemAt(position)
        data = item.data(0, Qt.UserRole) if item else None

        if data in [None, "connections_root"]:
            menu.addAction("Add Connection").triggered.connect(self.add_connection)
        if data in [None, "dbc_root"]:
            menu.addAction("Add Symbol File...").triggered.connect(self.add_dbc)
        if data in [None, "filter_root"]:
            menu.addAction("Add Filter").triggered.connect(self.add_filter)
        if data in [None, "canopen_root"]:
            menu.addAction("Add Node from EDS/DCF...").triggered.connect(
                self.add_canopen_node
            )
        if item and item.parent():
            menu.addAction("Remove").triggered.connect(lambda: self.remove_item(item))
        if menu.actions():
            menu.exec(self.tree.viewport().mapToGlobal(position))

    def add_connection(self):
        self.project.connections.append(
            Connection(
                name=f"Connection {len(self.project.connections) + 1}",
                config={"channel": f"vcan{len(self.project.connections)}"},
            )
        )
        self.rebuild_tree()

    def add_dbc(self):
        fns, _ = QFileDialog.getOpenFileNames(
            self,
            "Select DBC File(s)",
            "",
            "DBC, KCD, SYM, ARXML 3&4 and CDD Files (*.dbc *.arxml *.kcd *.sym *.cdd);;All Files (*)",
        )
        if fns:
            for fn in fns:
                try:
                    self.project.dbcs.append(
                        DBCFile(Path(fn), cantools.database.load_file(fn))
                    )
                except Exception as e:
                    QMessageBox.critical(
                        self, "DBC Load Error", f"Failed to load {Path(fn).name}: {e}"
                    )
            self.rebuild_tree()

    def add_filter(self):
        self.project.filters.append(
            CANFrameFilter(name=f"Filter {len(self.project.filters) + 1}")
        )
        self.rebuild_tree()

    def remove_item(self, item):
        if data := item.data(0, Qt.UserRole):
            if isinstance(data, DBCFile):
                self.project.dbcs.remove(data)
            elif isinstance(data, CANFrameFilter):
                self.project.filters.remove(data)
            elif isinstance(data, CANopenNode):
                self.project.canopen_nodes.remove(data)
            elif isinstance(data, Connection):
                removed_conn_id = data.id
                self.project.connections.remove(data)

                # Clean up references in other project items
                for dbc in self.project.dbcs:
                    if dbc.connection_id == removed_conn_id:
                        dbc.connection_id = -1
                for filt in self.project.filters:
                    if filt.connection_id == removed_conn_id:
                        filt.connection_id = -1
                for node in self.project.canopen_nodes:
                    if node.connection_id == removed_conn_id:
                        node.connection_id = -1

                # Clean up CANAsyncReader if it exists
                if removed_conn_id in self.main_window.can_readers:
                    reader = self.main_window.can_readers.pop(removed_conn_id)
                    reader.stop_reading()
                    reader.deleteLater()
                self.main_window.transmit_panel.set_connections(
                    self.main_window.can_readers
                )
            self.rebuild_tree()

    def add_canopen_node(self):
        fns, _ = QFileDialog.getOpenFileNames(
            self,
            "Select EDS/DCF File(s)",
            "",
            "CANopen Object Dictionary (*.eds *.dcf);;All Files (*)",
        )
        if fns:
            default_connection_id = None
            if self.project.connections:
                default_connection_id = self.project.connections[0].id

            for fn in fns:
                self.project.canopen_nodes.append(
                    CANopenNode(
                        path=Path(fn), node_id=1, connection_id=default_connection_id
                    )
                )
            self.rebuild_tree()


class TransmitViewColumn(enum.IntEnum):
    """Defines the columns for the TransmitPanel."""

    ON = 0
    ID = 1
    TYPE = 2
    RTR = 3
    DLC = 4
    DATA = 5
    CYCLE = 6
    SEND = 7
    SENT = 8


class TransmitPanel(QWidget):
    frame_to_send = Signal(object, object)  # message, connection_id (uuid.UUID)
    row_selection_changed = Signal(int, str, str)  # row, id_text, data_hex
    config_changed = Signal()

    def __init__(self):
        super().__init__()
        self.timers: Dict[int, QTimer] = {}
        self.dbcs: List[object] = []
        self.connections: Dict[str, CANAsyncReader] = {}
        self.setup_ui()

    def set_dbc_databases(self, dbs):
        self.dbcs = dbs

    def set_connections(self, connections: Dict[uuid.UUID, CANAsyncReader]):
        """Updates the connection list in the combo box, preserving selection."""
        self.connections = connections
        current_id = self.connection_combo.currentData()
        self.connection_combo.clear()

        if connections:
            sorted_connections = sorted(
                connections.values(), key=lambda r: r.connection.name
            )
            for reader in sorted_connections:
                self.connection_combo.addItem(
                    reader.connection.name, reader.connection.id
                )

            # Restore selection if possible
            if current_id and current_id in connections:
                index = self.connection_combo.findData(current_id)
                if index != -1:
                    self.connection_combo.setCurrentIndex(index)

            self.connection_combo.setEnabled(True)
        else:
            self.connection_combo.setEnabled(False)

    def get_message_from_id(self, can_id):
        for db_file in self.dbcs:
            try:
                return db_file.database.get_message_by_frame_id(can_id)
            except KeyError:
                continue

    def setup_ui(self):
        layout = QVBoxLayout(self)
        ctrl_layout = QHBoxLayout()
        self.add_btn = QPushButton("Add")
        self.rem_btn = QPushButton("Remove")
        self.connection_combo = QComboBox()
        self.connection_combo.setEnabled(False)
        self.connection_combo.setToolTip("Select the connection for sending frames")

        ctrl_layout.addWidget(self.add_btn)
        ctrl_layout.addWidget(self.rem_btn)
        ctrl_layout.addStretch()
        ctrl_layout.addWidget(QLabel("Send on:"))
        ctrl_layout.addWidget(self.connection_combo)
        layout.addLayout(ctrl_layout)

        self.table = QTableWidget()
        self.table.setColumnCount(9)
        headers = [col.name.replace("_", " ").title() for col in TransmitViewColumn]
        headers[TransmitViewColumn.ID] = "ID(hex)"
        headers[TransmitViewColumn.DATA] = "Data(hex)"
        self.table.setHorizontalHeaderLabels(headers)
        self.table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.setSelectionBehavior(QTableWidget.SelectRows)
        layout.addWidget(self.table)
        self.add_btn.clicked.connect(self.add_frame)
        self.rem_btn.clicked.connect(self.remove_frames)
        self.table.currentItemChanged.connect(self._on_item_changed)
        self.table.cellChanged.connect(self._on_cell_changed)

    def add_frame(self):
        r = self.table.rowCount()
        self.table.insertRow(r)
        self._setup_row_widgets(r)
        self.config_changed.emit()

    def add_message_frame(self, message: cantools.db.Message):
        """Adds a new row to the transmit table based on a DBC message definition."""
        r = self.table.rowCount()
        self.table.insertRow(r)
        self._setup_row_widgets(r)

        # Now, populate the new row with data from the message object
        self.table.blockSignals(True)

        # ID
        self.table.item(r, 1).setText(f"{message.frame_id:X}")

        # Type (Std/Ext)
        self.table.cellWidget(r, 2).setCurrentIndex(
            1 if message.is_extended_frame else 0
        )

        # DLC
        self.table.item(r, 4).setText(str(message.length))

        # Data (encode with defaults to get initial values)
        try:
            initial_data = message.encode({}, scaling=False, padding=True)
            self.table.item(r, 5).setText(initial_data.hex(" "))
        except Exception as e:
            print(f"Could not encode initial data for {message.name}: {e}")
            self.table.item(r, 5).setText("00 " * message.length)

        self.table.blockSignals(False)
        self.config_changed.emit()  # Mark project as dirty

    def remove_frames(self):
        if self.table.selectionModel().selectedRows():
            [
                self.table.removeRow(r)
                for r in sorted(
                    [i.row() for i in self.table.selectionModel().selectedRows()],
                    reverse=True,
                )
            ]
            self.config_changed.emit()

    def _setup_row_widgets(self, r):
        # ID
        self.table.setItem(r, TransmitViewColumn.ID, QTableWidgetItem("100"))

        # TYPE
        combo = QComboBox()
        combo.addItems(["Std", "Ext"])
        self.table.setCellWidget(r, TransmitViewColumn.TYPE, combo)
        combo.currentIndexChanged.connect(self.config_changed.emit)

        # RTR
        cb_rtr = QCheckBox()
        self.table.setCellWidget(r, TransmitViewColumn.RTR, self._center(cb_rtr))
        cb_rtr.toggled.connect(self.config_changed.emit)

        # DLC
        self.table.setItem(r, TransmitViewColumn.DLC, QTableWidgetItem("0"))

        # DATA
        self.table.setItem(r, TransmitViewColumn.DATA, QTableWidgetItem(""))

        # CYCLE
        self.table.setItem(r, TransmitViewColumn.CYCLE, QTableWidgetItem("100"))

        # SEND
        btn = QPushButton("Send")
        btn.clicked.connect(partial(self.send_from_row, r))
        self.table.setCellWidget(r, TransmitViewColumn.SEND, btn)

        # SENT
        sent_item = QTableWidgetItem("0")
        sent_item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
        self.table.setItem(r, TransmitViewColumn.SENT, sent_item)

        # ON
        cb_on = QCheckBox()
        cb_on.toggled.connect(partial(self._toggle_periodic, r))
        self.table.setCellWidget(r, TransmitViewColumn.ON, self._center(cb_on))

    def _center(self, w):
        c = QWidget()
        layout = QHBoxLayout(c)
        layout.addWidget(w)
        layout.setAlignment(Qt.AlignCenter)
        layout.setContentsMargins(0, 0, 0, 0)
        return c

    def _on_item_changed(self, curr, prev):
        if curr and (not prev or curr.row() != prev.row()):
            row = curr.row()
            id_text = self.table.item(row, TransmitViewColumn.ID).text()
            data_item = self.table.item(row, TransmitViewColumn.DATA)
            data_hex = data_item.text() if data_item else ""
            self.row_selection_changed.emit(row, id_text, data_hex)

    def _on_cell_changed(self, r, c):
        self.config_changed.emit()
        if c == TransmitViewColumn.ID or c == TransmitViewColumn.DATA:
            id_text = self.table.item(r, TransmitViewColumn.ID).text()
            data_item = self.table.item(r, TransmitViewColumn.DATA)
            data_hex = data_item.text() if data_item else ""
            self.row_selection_changed.emit(r, id_text, data_hex)
        elif c == TransmitViewColumn.DATA:
            self._update_dlc(r)

    def _update_dlc(self, r):
        try:
            data_len = len(
                bytes.fromhex(
                    self.table.item(r, TransmitViewColumn.DATA).text().replace(" ", "")
                )
            )
            self.table.item(r, TransmitViewColumn.DLC).setText(str(data_len))
        except (ValueError, TypeError):
            pass

    def update_row_data(self, r, data):
        self.table.blockSignals(True)
        self.table.item(r, TransmitViewColumn.DATA).setText(data.hex(" "))
        self.table.item(r, TransmitViewColumn.DLC).setText(str(len(data)))
        self.table.blockSignals(False)
        self.config_changed.emit()

    def _toggle_periodic(self, r, state):
        self.config_changed.emit()
        if state:
            try:
                cycle = int(self.table.item(r, TransmitViewColumn.CYCLE).text())
                t = QTimer(self)
                t.timeout.connect(partial(self.send_from_row, r))
                t.start(cycle)
                self.timers[r] = t
            except (ValueError, TypeError):
                QMessageBox.warning(self, "Bad Cycle", f"Row {r + 1}: bad cycle time.")
                self.table.cellWidget(r, TransmitViewColumn.ON).findChild(
                    QCheckBox
                ).setChecked(False)
        elif r in self.timers:
            self.timers.pop(r).stop()

    def stop_all_timers(self):
        [t.stop() for t in self.timers.values()]
        self.timers.clear()
        [
            self.table.cellWidget(r, 0).findChild(QCheckBox).setChecked(False)
            for r in range(self.table.rowCount())
        ]

    def send_from_row(self, r):
        connection_id = self.connection_combo.currentData()
        if not connection_id:
            QMessageBox.warning(
                self, "No Connection", "Please select a connection to send from."
            )
            return
        try:
            message_to_send = can.Message(
                arbitration_id=int(
                    self.table.item(r, TransmitViewColumn.ID).text(), 16
                ),
                is_extended_id=self.table.cellWidget(
                    r, TransmitViewColumn.TYPE
                ).currentIndex()
                == 1,
                is_remote_frame=self.table.cellWidget(r, TransmitViewColumn.RTR)
                .findChild(QCheckBox)
                .isChecked(),
                dlc=int(self.table.item(r, TransmitViewColumn.DLC).text()),
                data=bytes.fromhex(
                    self.table.item(r, TransmitViewColumn.DATA).text().replace(" ", "")
                ),
            )
            self.frame_to_send.emit(message_to_send, connection_id)

            sent_item = self.table.item(r, TransmitViewColumn.SENT)
            current_count = int(sent_item.text())
            sent_item.setText(str(current_count + 1))

        except (ValueError, TypeError) as e:
            QMessageBox.warning(self, "Bad Tx Data", f"Row {r + 1}: {e}")
            self._toggle_periodic(r, False)

    def send_selected(self):
        [
            self.send_from_row(r)
            for r in sorted(
                {i.row() for i in self.table.selectionModel().selectedIndexes()}
            )
        ]

    def get_config(self) -> List[Dict]:
        return [
            {
                "on": self.table.cellWidget(r, TransmitViewColumn.ON)
                .findChild(QCheckBox)
                .isChecked(),
                "id": self.table.item(r, TransmitViewColumn.ID).text(),
                "type_idx": self.table.cellWidget(
                    r, TransmitViewColumn.TYPE
                ).currentIndex(),
                "rtr": self.table.cellWidget(r, TransmitViewColumn.RTR)
                .findChild(QCheckBox)
                .isChecked(),
                "dlc": self.table.item(r, TransmitViewColumn.DLC).text(),
                "data": self.table.item(r, TransmitViewColumn.DATA).text(),
                "cycle": self.table.item(r, TransmitViewColumn.CYCLE).text(),
            }
            for r in range(self.table.rowCount())
        ]

    def set_config(self, config: List[Dict]):
        self.stop_all_timers()
        self.table.clearContents()
        self.table.setRowCount(0)
        self.table.setRowCount(len(config))
        self.table.blockSignals(True)
        for r, row_data in enumerate(config):
            self._setup_row_widgets(r)
            self.table.cellWidget(r, TransmitViewColumn.ON).findChild(
                QCheckBox
            ).setChecked(row_data.get("on", False))
            self.table.item(r, TransmitViewColumn.ID).setText(row_data.get("id", "0"))
            self.table.cellWidget(r, TransmitViewColumn.TYPE).setCurrentIndex(
                row_data.get("type_idx", 0)
            )
            self.table.cellWidget(r, TransmitViewColumn.RTR).findChild(
                QCheckBox
            ).setChecked(row_data.get("rtr", False))
            self.table.item(r, TransmitViewColumn.DLC).setText(row_data.get("dlc", "0"))
            self.table.item(r, TransmitViewColumn.DATA).setText(
                row_data.get("data", "")
            )
            self.table.item(r, TransmitViewColumn.CYCLE).setText(
                row_data.get("cycle", "100")
            )
        self.table.blockSignals(False)
        self.config_changed.emit()


class SignalTransmitPanel(QGroupBox):
    data_encoded = Signal(bytes)

    def __init__(self):
        super().__init__("Signal Config")
        self.message = None
        self.setup_ui()

    def setup_ui(self):
        layout = QVBoxLayout(self)
        self.table = QTableWidget()
        self.table.setColumnCount(6)
        self.table.setHorizontalHeaderLabels(
            ["Signal", "Value", "Unit", "Min", "Max", "Status"]
        )
        self.table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        self.table.horizontalHeader().setStretchLastSection(True)
        layout.addWidget(self.table)
        self.table.cellChanged.connect(self._validate_and_encode)

    def clear_panel(self):
        self.message = None
        self.table.setRowCount(0)
        self.setTitle("Signal Config")
        self.setVisible(False)

    def _set_status(self, row: int, text: str, is_error: bool):
        """Helper to set the text and color of a status cell."""
        status_item = QTableWidgetItem(text)
        if is_error:
            status_item.setForeground(Qt.red)
            status_item.setToolTip(text)
        else:
            status_item.setForeground(Qt.green)

        status_item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
        self.table.setItem(row, 5, status_item)

    def populate(self, msg, data_hex: str):
        self.message = msg
        initial_values = {}

        # Try to decode the existing data to pre-fill the values
        if data_hex:
            try:
                data_bytes = bytes.fromhex(data_hex.replace(" ", ""))
                # Use allow_truncated=True to handle cases where data is shorter than expected
                initial_values = msg.decode(
                    data_bytes, decode_choices=False, allow_truncated=True
                )
            except (ValueError, KeyError) as e:
                print(f"Could not decode existing data for signal panel: {e}")
                initial_values = {}  # Fallback to defaults on error

        self.table.blockSignals(True)
        self.table.setRowCount(len(msg.signals))

        for r, s in enumerate(msg.signals):
            # Use the decoded value if available, otherwise use the signal's default initial value
            value = initial_values.get(
                s.name, s.initial if s.initial is not None else 0
            )

            self.table.setItem(r, 0, QTableWidgetItem(s.name))
            self.table.item(r, 0).setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)

            self.table.setItem(r, 1, QTableWidgetItem(str(value)))

            self.table.setItem(r, 2, QTableWidgetItem(str(s.unit or "")))
            self.table.item(r, 2).setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)

            # --- Min (Col 3) - NEW ---
            min_val = s.minimum if s.minimum is not None else "N/A"
            min_item = QTableWidgetItem(str(min_val))
            min_item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
            min_item.setToolTip(f"Minimum allowed value for {s.name}")
            self.table.setItem(r, 3, min_item)

            # --- Max (Col 4) - NEW ---
            max_val = s.maximum if s.maximum is not None else "N/A"
            max_item = QTableWidgetItem(str(max_val))
            max_item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
            max_item.setToolTip(f"Maximum allowed value for {s.name}")
            self.table.setItem(r, 4, max_item)

            self.table.setItem(r, 5, QTableWidgetItem(""))

        self.table.blockSignals(False)
        self.setTitle(f"Signal Config: {msg.name}")
        self.setVisible(True)
        # Trigger an initial encode to ensure the data field is up-to-date,
        # especially if we fell back to default values.
        self._validate_and_encode()

    def _validate_and_encode(self):
        """
        Validates the signal values against the DBC, updates the status column
        with any errors, and emits the encoded data if successful.
        """
        if not self.message:
            return

        # --- 1. Build the data dictionary from the table ---
        data_dict = {}
        parse_errors = False

        # We don't need to block signals here as we are just reading
        for r in range(self.table.rowCount()):
            signal_name = self.table.item(r, 0).text()
            value_text = self.table.item(r, 1).text()
            try:
                data_dict[signal_name] = float(value_text)
            except (ValueError, TypeError):
                parse_errors = True
                # We will set the status message for this *after* reading all values

        # --- Temporarily block signals to prevent recursion when updating status ---
        self.table.blockSignals(True)

        # Update status based on initial parsing
        for r in range(self.table.rowCount()):
            signal_name = self.table.item(r, 0).text()
            if signal_name not in data_dict:
                self._set_status(r, "Invalid number", is_error=True)
            else:
                self._set_status(r, "", is_error=False)  # Clear previous parse errors

        self.table.blockSignals(False)
        # --------------------------------------------------------------------------

        if parse_errors:
            return

        # --- 2. Attempt to encode and handle validation errors ---
        try:
            encoded_data = self.message.encode(data_dict, strict=True)

            # Block signals again for the success case
            self.table.blockSignals(True)
            for r in range(self.table.rowCount()):
                self._set_status(r, "OK", is_error=False)
            self.table.blockSignals(False)

            self.data_encoded.emit(encoded_data)

        except (cantools.database.errors.EncodeError, ValueError, KeyError) as e:
            error_str = str(e)

            # Block signals while we update rows with error messages
            self.table.blockSignals(True)
            found_error_signal = False
            for r in range(self.table.rowCount()):
                signal_name = self.table.item(r, 0).text()
                if f'"{signal_name}"' in error_str:
                    self._set_status(r, error_str, is_error=True)
                    found_error_signal = True
                else:
                    self._set_status(r, "OK", is_error=False)

            if not found_error_signal:
                self._set_status(0, error_str, is_error=True)

            self.table.blockSignals(False)


# --- Main Application Window ---
class CANBusObserver(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("CANPeek")
        self.setGeometry(100, 100, 1400, 900)
        self.canopen_network = None  # Will be created when connecting
        self.MAX_RECENT_PROJECTS = 10
        self.recent_projects_paths = []
        self.interface_manager = CANInterfaceManager()
        self.pdo_manager = PDODatabaseManager()
        self.project = Project()
        self.current_project_path: Optional[Path] = None
        self.project_dirty = False
        self.can_readers: Dict[uuid.UUID, CANAsyncReader] = {}
        self.bus_states: Dict[uuid.UUID, can.BusState] = {}

        file_loggers = {
            "ASCWriter": ".asc",
            "BLFWriter": ".blf",
            "CSVWriter": ".csv",
            "SqliteWriter": ".db",
            "CanutilsLogWriter": ".log",
            "TRCWriter": ".trc",
            "Printer": ".txt",
        }
        sorted_loggers = sorted(file_loggers.items())
        filters = [f"{ext} : {name} Log (*{ext})" for name, ext in sorted_loggers]
        filters += [
            f"{ext}.gz : Compressed {name} Log (*{ext}.gz)"
            for name, ext in sorted_loggers
        ]
        self.log_file_filter = ";;".join(filters)
        self.log_file_filter_open = (
            f"All Supported ({' '.join(['*' + ext for _, ext in sorted_loggers])});;"
            + self.log_file_filter
        )
        self.trace_model = CANTraceModel()
        self.grouped_model = CANGroupedModel()
        self.grouped_proxy_model = QSortFilterProxyModel()
        self.grouped_proxy_model.setSourceModel(self.grouped_model)
        self.grouped_proxy_model.setSortRole(Qt.UserRole)
        self.frame_batch = []
        self.all_received_frames = []

        QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.OpaqueSplitterResize, True)
        QtAds.CDockManager.setConfigFlag(
            QtAds.CDockManager.XmlCompressionEnabled, False
        )
        QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.FocusHighlighting, True)
        self.dock_manager = QtAds.CDockManager(self)

        # Set monospaced font for grouped and trace views
        monospace_font = QFont("monospace")
        monospace_font.setStyleHint(QFont.Monospace)

        self.grouped_view = QTreeView()
        self.grouped_view.setModel(self.grouped_proxy_model)
        self.grouped_view.setAlternatingRowColors(True)
        self.grouped_view.setSortingEnabled(True)
        self.grouped_view.setFont(monospace_font)

        self.trace_view_widget = QWidget()
        trace_layout = QVBoxLayout(self.trace_view_widget)
        trace_layout.setContentsMargins(5, 5, 5, 5)
        self.trace_view = QTableView()
        self.trace_view.setModel(self.trace_model)
        self.trace_view.setAlternatingRowColors(True)
        self.trace_view.horizontalHeader().setStretchLastSection(True)
        self.trace_view.setFont(monospace_font)
        self.autoscroll_cb = QCheckBox("Autoscroll", checked=True)
        trace_layout.addWidget(self.trace_view)
        trace_layout.addWidget(self.autoscroll_cb)

        self.object_dictionary_viewer = ObjectDictionaryViewer()
        self.object_dictionary_viewer.frame_to_send.connect(self.send_can_frame)

        self.nmt_sender = NMTSender(self.project)
        self.nmt_sender.frame_to_send.connect(self.send_can_frame)
        self.nmt_sender.status_update.connect(self.statusBar().showMessage)

        # self.setCentralWidget(self.dock_manager)

        self.setup_actions()
        self.setup_ui()
        self.setup_docks()
        self.setup_toolbar()
        self.setup_menubar()
        self.setup_statusbar()
        self._load_recent_projects()
        self._update_recent_projects_menu()
        self.project_explorer.project_changed.connect(lambda: self._set_dirty(True))
        self.transmit_panel.config_changed.connect(lambda: self._set_dirty(True))
        self.restore_layout()
        self.update_perspectives_menu()
        self.gui_update_timer = QTimer(self)
        self.gui_update_timer.timeout.connect(self.update_views)
        self.gui_update_timer.start(50)
        self._update_window_title()

        icon = QIcon(QPixmap(":/icons/canpeek.png"))
        self.setWindowIcon(icon)

        self.project_explorer.project_changed.connect(
            self._on_project_structure_changed
        )

        # Apply default theme and style
        settings = QSettings()
        saved_theme = settings.value("selectedTheme", "default")
        self._set_theme(saved_theme)

        # Check the corresponding action in the menu
        for action in self.theme_group.actions():
            if action.text() == saved_theme:
                action.setChecked(True)
                break

    def _set_theme(self, theme_name: str):
        if theme_name == "default":
            theme_name = None

        # TODO : theme is partialy applied if run once ....
        qt_themes.set_theme(theme_name)
        qt_themes.set_theme(theme_name)
        qt_themes.set_theme(theme_name)
        settings = QSettings()
        settings.setValue("selectedTheme", theme_name)

    def setup_actions(self):
        style = self.style()
        self.new_project_action = QAction(
            style.standardIcon(QStyle.SP_FileIcon), "&New Project", self
        )
        self.open_project_action = QAction(
            style.standardIcon(QStyle.SP_DialogOpenButton), "&Open Project...", self
        )
        self.save_project_action = QAction(
            style.standardIcon(QStyle.SP_DialogSaveButton), "Save &Project", self
        )
        self.save_project_as_action = QAction(
            QIcon(QPixmap(":/icons/document-save-as.png")), "Save Project &As...", self
        )
        self.connect_action = QAction(
            style.standardIcon(QStyle.SP_DialogYesButton), "&Connect", self
        )
        self.disconnect_action = QAction(
            style.standardIcon(QStyle.SP_DialogNoButton), "&Disconnect", self
        )
        self.clear_action = QAction(
            style.standardIcon(QStyle.SP_TrashIcon), "&Clear Data", self
        )
        self.save_log_action = QAction(
            QIcon(QPixmap(":/icons/document-export.png")), "&Save Log...", self
        )
        self.load_log_action = QAction(
            QIcon(QPixmap(":/icons/document-import.png")), "&Load Log...", self
        )
        self.exit_action = QAction("&Exit", self)
        self.new_project_action.triggered.connect(self._new_project)
        self.open_project_action.triggered.connect(self._open_project)
        self.save_project_action.triggered.connect(self._save_project)
        self.save_project_as_action.triggered.connect(self._save_project_as)
        self.connect_action.triggered.connect(
            lambda: asyncio.create_task(self.connect_can())
        )
        self.disconnect_action.triggered.connect(self.disconnect_can)
        self.clear_action.triggered.connect(self.clear_data)
        self.save_log_action.triggered.connect(self.save_log)
        self.load_log_action.triggered.connect(self.load_log)
        self.exit_action.triggered.connect(self.close)
        self.disconnect_action.setEnabled(False)

    def setup_toolbar(self):
        toolbar = QToolBar("Main Toolbar")
        toolbar.setObjectName("MainToolbar")
        self.addToolBar(toolbar)
        toolbar.addAction(self.new_project_action)
        toolbar.addAction(self.open_project_action)
        toolbar.addAction(self.save_project_action)
        toolbar.addSeparator()
        toolbar.addAction(self.connect_action)
        toolbar.addAction(self.disconnect_action)
        toolbar.addSeparator()
        toolbar.addAction(self.clear_action)
        toolbar.addAction(self.save_log_action)
        toolbar.addAction(self.load_log_action)

    def setup_ui(self):
        # Set monospaced font for grouped and trace views
        monospace_font = QFontDatabase.systemFont(QFontDatabase.FixedFont)

        self.grouped_view = QTreeView()
        self.grouped_view.setModel(self.grouped_proxy_model)
        self.grouped_view.setAlternatingRowColors(True)
        self.grouped_view.setSortingEnabled(True)
        self.grouped_view.setFont(monospace_font)
        self.grouped_view.sortByColumn(GroupedViewColumn.ID, Qt.AscendingOrder)

        self.trace_view_widget = QWidget()
        trace_layout = QVBoxLayout(self.trace_view_widget)
        trace_layout.setContentsMargins(5, 5, 5, 5)
        self.trace_view = QTableView()
        self.trace_view.setModel(self.trace_model)
        self.trace_view.setAlternatingRowColors(True)
        self.trace_view.horizontalHeader().setStretchLastSection(True)
        self.trace_view.setFont(monospace_font)
        self.autoscroll_cb = QCheckBox("Autoscroll", checked=True)
        trace_layout.addWidget(self.trace_view)
        trace_layout.addWidget(self.autoscroll_cb)

        self.object_dictionary_viewer = ObjectDictionaryViewer()
        self.object_dictionary_viewer.frame_to_send.connect(self.send_can_frame)

        self.nmt_sender = NMTSender(self.project)
        self.nmt_sender.frame_to_send.connect(self.send_can_frame)
        self.nmt_sender.status_update.connect(self.statusBar().showMessage)

    def setup_docks(self):
        # Set central widget
        label = QLabel()
        label.setText(
            "This is a DockArea which is always visible, even if it does not contain any DockWidgets."
        )
        label.setAlignment(Qt.AlignCenter)
        central_dock_widget = QtAds.CDockWidget("CentralWidget")
        central_dock_widget.setWidget(label)
        central_dock_widget.setFeature(QtAds.CDockWidget.NoTab, True)
        central_dock_area = self.dock_manager.setCentralWidget(central_dock_widget)

        self.project_explorer = ProjectExplorer(self.project, self)
        explorer_dock = QtAds.CDockWidget("Project Explorer")
        explorer_dock.setWidget(self.project_explorer)
        project_aera = self.dock_manager.addDockWidget(
            QtAds.DockWidgetArea.LeftDockWidgetArea, explorer_dock
        )

        self.properties_panel = PropertiesPanel(
            self.project, self.project_explorer, self.interface_manager, self
        )
        properties_dock = QtAds.CDockWidget("Properties")
        properties_dock.setWidget(self.properties_panel)
        self.dock_manager.addDockWidget(
            QtAds.DockWidgetArea.BottomDockWidgetArea, properties_dock, project_aera
        )

        # New docks for previously tabbed views
        grouped_dock = QtAds.CDockWidget("Grouped View")
        grouped_dock.setWidget(self.grouped_view)
        self.dock_manager.addDockWidgetTabToArea(grouped_dock, central_dock_area)

        trace_dock = QtAds.CDockWidget("Trace View")
        trace_dock.setWidget(self.trace_view_widget)
        self.dock_manager.addDockWidgetTabToArea(trace_dock, central_dock_area)

        object_dictionary_dock = QtAds.CDockWidget("Object Dictionary")
        object_dictionary_dock.setWidget(self.object_dictionary_viewer)
        self.dock_manager.addDockWidgetTabToArea(
            object_dictionary_dock, central_dock_area
        )
        # Set the Grouped View active
        central_dock_area.setCurrentDockWidget(grouped_dock)

        transmit_container = QWidget()
        transmit_layout = QVBoxLayout(transmit_container)
        transmit_layout.setContentsMargins(0, 0, 0, 0)
        self.transmit_panel = TransmitPanel()
        self.signal_transmit_panel = SignalTransmitPanel()
        transmit_layout.addWidget(self.transmit_panel)
        transmit_layout.addWidget(self.signal_transmit_panel)
        self.signal_transmit_panel.setVisible(False)
        transmit_dock = QtAds.CDockWidget("Transmit")
        transmit_dock.setWidget(transmit_container)
        transmit_aera = self.dock_manager.addDockWidget(
            QtAds.BottomDockWidgetArea, transmit_dock, central_dock_area
        )

        nmt_sender_dock = QtAds.CDockWidget("NMT Sender")
        nmt_sender_dock.setWidget(self.nmt_sender)
        self.dock_manager.addDockWidgetTabToArea(nmt_sender_dock, transmit_aera)

        transmit_aera.setCurrentDockWidget(transmit_dock)

        transmit_aera

        self.docks = {
            "explorer": explorer_dock,
            "properties": properties_dock,
            "transmit": transmit_dock,
            "grouped": grouped_dock,
            "trace": trace_dock,
            "object_dictionary": object_dictionary_dock,
            "nmt_sender": nmt_sender_dock,
        }

        self.properties_panel.message_to_transmit.connect(
            self._add_message_to_transmit_panel
        )
        self.transmit_panel.frame_to_send.connect(self.send_can_frame)
        self.transmit_panel.row_selection_changed.connect(self.on_transmit_row_selected)
        self.signal_transmit_panel.data_encoded.connect(self.on_signal_data_encoded)
        self.project_explorer.project_changed.connect(self.on_project_changed)
        self.project_explorer.project_changed.connect(
            self.nmt_sender.update_project_nodes
        )
        self.project_explorer.tree.currentItemChanged.connect(
            self.properties_panel.show_properties
        )
        self.project_explorer.tree.currentItemChanged.connect(
            self.on_project_explorer_selection_changed
        )

    def save_perspective(self):
        perspective_name, ok = QInputDialog.getText(
            self, "Save Perspective", "Enter a name for the perspective:"
        )
        if ok and perspective_name:
            self.dock_manager.addPerspective(perspective_name)
            self.update_perspectives_menu()

    def load_perspective(self, index):
        perspective_name = self.perspectives_combobox.itemText(index)
        self.dock_manager.openPerspective(perspective_name)

    def update_perspectives_menu(self):
        self.perspectives_combobox.clear()
        self.perspectives_combobox.addItems(self.dock_manager.perspectiveNames())

    def setup_menubar(self):
        menubar = self.menuBar()
        file_menu = menubar.addMenu("&File")
        file_menu.addAction(self.new_project_action)
        file_menu.addAction(self.open_project_action)
        self.recent_menu = QMenu("Open &Recent", self)
        self.recent_menu.setIcon(QIcon(QPixmap(":/icons/document-open-recent.png")))
        file_menu.addMenu(self.recent_menu)
        file_menu.addAction(self.save_project_action)
        file_menu.addAction(self.save_project_as_action)
        file_menu.addSeparator()
        file_menu.addAction(self.clear_action)
        file_menu.addAction(self.load_log_action)
        file_menu.addAction(self.save_log_action)
        file_menu.addSeparator()
        file_menu.addAction(self.exit_action)
        connect_menu = menubar.addMenu("&Connect")
        connect_menu.addAction(self.connect_action)
        connect_menu.addAction(self.disconnect_action)
        view_menu = menubar.addMenu("&View")
        for dock in self.docks.values():
            view_menu.addAction(dock.toggleViewAction())

        perspectives_menu = menubar.addMenu("&Perspectives")
        save_perspective_action = QAction("Save Perspective...", self)
        save_perspective_action.triggered.connect(self.save_perspective)
        perspectives_menu.addAction(save_perspective_action)

        remove_perspective_action = QAction("Remove Perspective...", self)
        remove_perspective_action.triggered.connect(self._remove_perspective)
        perspectives_menu.addAction(remove_perspective_action)

        self.perspectives_combobox = QComboBox(self)
        self.perspectives_combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.perspectives_combobox.setSizePolicy(
            QSizePolicy.Preferred, QSizePolicy.Preferred
        )
        self.perspectives_combobox.activated.connect(self.load_perspective)

        # Theme and Style menus
        theme_menu = menubar.addMenu("&Themes")
        self.theme_group = QActionGroup(self)
        self.theme_group.setExclusive(True)
        names = ["default"]
        names.extend(sorted(qt_themes.get_themes().keys()))
        for theme_name in names:
            action = QAction(theme_name, self, checkable=True)
            action.triggered.connect(partial(self._set_theme, theme_name))
            self.theme_group.addAction(action)
            theme_menu.addAction(action)

        perspective_list_action = QWidgetAction(self)
        perspective_list_action.setDefaultWidget(self.perspectives_combobox)
        perspectives_menu.addAction(perspective_list_action)

    def setup_statusbar(self):
        self.statusBar().showMessage("Ready")
        self.frame_count_label = QLabel("Frames: 0")
        self.connection_label = QLabel("Disconnected")
        self.bus_state_label = QLabel("Bus States: N/A")
        self.statusBar().addPermanentWidget(self.bus_state_label)
        self.statusBar().addPermanentWidget(self.frame_count_label)
        self.statusBar().addPermanentWidget(self.connection_label)

    def _remove_perspective(self):
        perspective_names = self.dock_manager.perspectiveNames()
        if not perspective_names:
            QMessageBox.information(
                self, "Remove Perspective", "No perspectives to remove."
            )
            return

        perspective_name, ok = QInputDialog.getItem(
            self,
            "Remove Perspective",
            "Select a perspective to remove:",
            perspective_names,
            0,
            False,
        )
        if ok and perspective_name:
            self.dock_manager.removePerspective(perspective_name)
            self.update_perspectives_menu()

    def _add_message_to_transmit_panel(self, message: cantools.db.Message):
        """Slot to handle request to add a DBC message to the transmit panel."""
        self.transmit_panel.add_message_frame(message)

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key_Space and self.transmit_panel.table.hasFocus():
            self.transmit_panel.send_selected()
            event.accept()
        else:
            super().keyPressEvent(event)

    def _process_frame(self, frame: CANFrame):
        try:
            self.frame_batch.append(frame)
            if frame.arbitration_id & 0x580 == 0x580:
                self.object_dictionary_viewer.frame_rx_sdo.emit(frame)
        except Exception as e:
            print(f"Error processing frame: {e}")

    def update_views(self):
        if not self.frame_batch:
            return
        try:
            frames_to_process, self.frame_batch = self.frame_batch[:], []
            active_filters = self.project.get_active_filters()
            filtered_frames = [
                f
                for f in frames_to_process
                if not active_filters or any(filt.matches(f) for filt in active_filters)
            ]
            if not filtered_frames:
                return
            expanded_ids = {
                self.grouped_model.data(
                    self.grouped_proxy_model.mapToSource(
                        self.grouped_proxy_model.index(row, 0)
                    ),
                    Qt.UserRole,
                )
                for row in range(self.grouped_proxy_model.rowCount())
                if self.grouped_view.isExpanded(self.grouped_proxy_model.index(row, 0))
            }
            self.grouped_model.update_frames(filtered_frames)
            self.all_received_frames.extend(filtered_frames)
            if len(self.all_received_frames) > TRACE_BUFFER_LIMIT:
                del self.all_received_frames[:-TRACE_BUFFER_LIMIT]
            self.trace_model.set_data(self.all_received_frames)
            for row in range(self.grouped_proxy_model.rowCount()):
                proxy_index = self.grouped_proxy_model.index(row, 0)
                if (
                    self.grouped_model.data(
                        self.grouped_proxy_model.mapToSource(proxy_index), Qt.UserRole
                    )
                    in expanded_ids
                ):
                    self.grouped_view.setExpanded(proxy_index, True)
            if self.autoscroll_cb.isChecked():
                self.trace_view.scrollToBottom()
            self.frame_count_label.setText(f"Frames: {len(self.all_received_frames)}")
        except Exception as e:
            import traceback

            print(f"Error in update_views: {e}")
            traceback.print_exc()

    def _create_pdo_databases(self) -> List[object]:
        """Create PDO databases from enabled CANopen nodes using manager"""
        return self.pdo_manager.get_all_active_databases(self.project.canopen_nodes)

    def on_project_changed(self):
        active_dbcs = self.project.get_active_dbcs()
        pdo_databases = self._create_pdo_databases()

        self.trace_model.set_config(
            active_dbcs, self.project.canopen_enabled, pdo_databases
        )
        self.grouped_model.set_config(
            active_dbcs, self.project.canopen_enabled, pdo_databases
        )

        if self.can_readers:
            asyncio.create_task(self._update_canopen_nodes())

        self.transmit_panel.set_connections(self.can_readers)
        self.transmit_panel.set_dbc_databases(active_dbcs)

        row = self.transmit_panel.table.currentRow()
        id_text = ""
        data_hex = ""
        if row >= 0:
            id_item = self.transmit_panel.table.item(row, 1)
            if id_item:
                id_text = id_item.text()

            data_item = self.transmit_panel.table.item(row, 5)
            if data_item:
                data_hex = data_item.text()

        self.on_transmit_row_selected(row, id_text, data_hex)

        self.properties_panel.project = self.project

        current_item = self.project_explorer.tree.currentItem()
        if current_item:
            self.on_project_explorer_selection_changed(current_item, None)
        else:
            self.object_dictionary_viewer.clear_node()

    def on_transmit_row_selected(self, row: int, id_text: str, data_hex: str):
        self.signal_transmit_panel.clear_panel()
        if row < 0 or not id_text:
            return
        try:
            can_id = int(id_text, 16)
            message = self.transmit_panel.get_message_from_id(can_id)
            if message:
                self.signal_transmit_panel.populate(message, data_hex)
        except ValueError:
            pass

    def on_signal_data_encoded(self, data_bytes):
        if (row := self.transmit_panel.table.currentRow()) >= 0:
            self.transmit_panel.update_row_data(row, data_bytes)

    async def _update_canopen_nodes(self):
        """Update CANopen nodes when project changes while connected"""
        if not self.canopen_network:
            return

        self.canopen_network.nodes.clear()

        if self.project.canopen_enabled:
            for node_config in self.project.canopen_nodes:
                if node_config.enabled and node_config.path.exists():
                    try:
                        await self.canopen_network.aadd_node(
                            node_config.node_id, str(node_config.path)
                        )
                        print(f"Updated CANopen node {node_config.node_id}")
                    except Exception as e:
                        print(f"Error updating CANopen node {node_config.node_id}: {e}")

    def on_project_explorer_selection_changed(self, current, previous):
        """Handle project explorer selection changes"""
        if not current:
            self.object_dictionary_viewer.clear_node()
            return

        is_connected = bool(self.can_readers)
        data = current.data(0, Qt.UserRole)
        if isinstance(data, CANopenNode) and data.enabled:
            asyncio.create_task(self.object_dictionary_viewer.set_node(data))
            self.nmt_sender.set_connection_context(data.connection_id, is_connected)
        elif isinstance(data, Connection):
            self.nmt_sender.set_connection_context(data.id, is_connected)
        elif isinstance(data, str) and data.startswith("canopen_bus_"):
            connection_id_str = data.replace("canopen_bus_", "")
            try:
                connection_id = uuid.UUID(connection_id_str)
                self.nmt_sender.set_connection_context(connection_id, is_connected)
            except ValueError:
                self.nmt_sender.set_connection_context(None, is_connected)
        else:
            self.object_dictionary_viewer.clear_node()
            self.nmt_sender.set_connection_context(None, is_connected)

    def _update_bus_state(self, state: can.BusState):
        """Updates the status bar with the current CAN bus state."""
        sender_reader = self.sender()
        if not isinstance(sender_reader, CANAsyncReader):
            return

        conn_id = sender_reader.connection.id
        self.bus_states[conn_id] = state

        state_strings = []
        for conn_id, s in sorted(
            self.bus_states.items(),
            key=lambda item: self.project.get_connection_name(item[0]),
        ):
            conn_name = self.project.get_connection_name(conn_id)
            state_strings.append(
                f"<span style='color: {self._get_state_color(s)};'>{conn_name}: {s.name.title()}</span>"
            )

        self.bus_state_label.setText("Bus States: " + ", ".join(state_strings))

    def _get_state_color(self, state: can.BusState) -> str:
        if state == can.BusState.ACTIVE:
            return "#4CAF50"  # Green
        elif state == can.BusState.PASSIVE:
            return "#FFC107"  # Amber
        elif state == can.BusState.ERROR:
            return "#F44336"  # Red
        return "#FFFFFF"  # White/Default

    async def connect_can(self):
        active_connections = self.project.get_active_connections()
        if not active_connections:
            QMessageBox.information(
                self, "No Connections", "No active connections to start."
            )
            return

        connect_tasks = [self._connect_single(conn) for conn in active_connections]
        results = await asyncio.gather(*connect_tasks)

        successful_connections = [
            conn for conn, res in zip(active_connections, results) if res
        ]

        if successful_connections:
            self.connect_action.setEnabled(False)
            self.disconnect_action.setEnabled(True)
            self.transmit_panel.set_connections(self.can_readers)

            connected_names = [conn.name for conn in successful_connections]
            self.connection_label.setText(f"Connected to: {', '.join(connected_names)}")
        else:
            self.connection_label.setText("Connection Failed")

        if current_item := self.project_explorer.tree.currentItem():
            self.properties_panel.show_properties(current_item)
            if isinstance(self.properties_panel.current_widget, ConnectionEditor):
                self.properties_panel.current_widget.set_connected_state(True)

        # Update NMT sender context after connection state changes
        self.on_project_explorer_selection_changed(
            self.project_explorer.tree.currentItem(), None
        )

    async def _connect_single(self, connection: Connection) -> bool:
        if connection.id in self.can_readers:
            return True  # Already connected

        reader = CANAsyncReader(connection)
        reader.frame_received.connect(self._process_frame)
        reader.error_occurred.connect(self.on_can_error)
        reader.bus_state_changed.connect(self._update_bus_state)

        if await reader.start_reading():
            self.can_readers[connection.id] = reader
            return True
        else:
            reader.deleteLater()
            return False

    def disconnect_can(self):
        for reader in self.can_readers.values():
            reader.stop_reading()
            reader.deleteLater()
        self.can_readers.clear()
        self.bus_states.clear()

        self.connect_action.setEnabled(True)
        self.disconnect_action.setEnabled(False)
        self.transmit_panel.stop_all_timers()
        self.transmit_panel.set_connections(self.can_readers)
        self.connection_label.setText("Disconnected")
        self.bus_state_label.setText("Bus States: N/A")

        if current_item := self.project_explorer.tree.currentItem():
            self.properties_panel.show_properties(current_item)
            if isinstance(self.properties_panel.current_widget, ConnectionEditor):
                self.properties_panel.current_widget.set_connected_state(False)

        # Update NMT sender context after connection state changes
        self.on_project_explorer_selection_changed(
            self.project_explorer.tree.currentItem(), None
        )

    def send_can_frame(self, message: can.Message, connection_id: uuid.UUID):
        print(f"Sending frame: {message} on connection {connection_id}")
        if reader := self.can_readers.get(connection_id):
            if reader.running:
                reader.send_frame(message)
        else:
            conn_name = self.project.get_connection_name(connection_id)
            QMessageBox.warning(
                self, "Not Connected", f"Connection '{conn_name}' is not active."
            )

    def on_can_error(self, error_message: str):
        QMessageBox.warning(self, "CAN Error", error_message)
        self.statusBar().showMessage(f"Error: {error_message}")
        self.disconnect_can()

    def clear_data(self):
        self.all_received_frames.clear()
        self.grouped_model.clear_frames()
        self.trace_model.set_data([])
        self.frame_count_label.setText("Frames: 0")

    def save_log(self):
        if not self.all_received_frames:
            QMessageBox.information(self, "No Data", "No frames to save.")
            return
        dialog = QFileDialog(self, "Save CAN Log", "", self.log_file_filter)
        dialog.setDefaultSuffix("log")
        dialog.setAcceptMode(QFileDialog.AcceptSave)
        if not dialog.exec():
            return
        filename = dialog.selectedFiles()[0]
        logger = None
        try:
            logger = can.Logger(filename)
            for frame in self.all_received_frames:
                logger.on_message_received(
                    can.Message(
                        timestamp=frame.timestamp,
                        arbitration_id=frame.arbitration_id,
                        is_extended_id=frame.is_extended,
                        is_remote_frame=frame.is_remote,
                        is_error_frame=frame.is_error,
                        dlc=frame.dlc,
                        data=frame.data,
                        channel=frame.channel,
                    )
                )
            self.statusBar().showMessage(f"Log saved to {filename}")
        except Exception as e:
            QMessageBox.critical(self, "Save Error", f"Failed to save log: {e}")
        finally:
            if logger:
                logger.stop()

    def load_log(self):
        filename, _ = QFileDialog.getOpenFileName(
            self, "Load CAN Log", "", self.log_file_filter_open
        )
        if not filename:
            return
        try:
            self.clear_data()
            frames_to_add = []
            for msg in can.LogReader(filename):
                frames_to_add.append(
                    CANFrame(
                        timestamp=msg.timestamp,
                        arbitration_id=msg.arbitration_id,
                        data=msg.data,
                        dlc=msg.dlc,
                        is_extended=msg.is_extended_id,
                        is_error=msg.is_error_frame,
                        is_remote=msg.is_remote_frame,
                        channel=msg.channel or "CAN1",
                    )
                )
            self.frame_batch.extend(frames_to_add)
            self.update_views()
            self.statusBar().showMessage(
                f"Loaded {len(self.all_received_frames)} frames from {filename}"
            )
        except Exception as e:
            QMessageBox.critical(self, "Load Error", f"Failed to load log: {e}")

    def _load_recent_projects(self):
        self.recent_projects_paths = QSettings().value("recentProjects", [], type=list)

    def _save_recent_projects(self):
        QSettings().setValue("recentProjects", self.recent_projects_paths)

    def _add_to_recent_projects(self, path: Path):
        path_str = str(path.resolve())
        if path_str in self.recent_projects_paths:
            self.recent_projects_paths.remove(path_str)
        self.recent_projects_paths.insert(0, path_str)
        self.recent_projects_paths = self.recent_projects_paths[
            : self.MAX_RECENT_PROJECTS
        ]
        self._update_recent_projects_menu()
        self._save_recent_projects()

    def _update_recent_projects_menu(self):
        self.recent_menu.clear()
        if not self.recent_projects_paths:
            self.recent_menu.addAction(
                QAction("No Recent Projects", self, enabled=False)
            )
            return
        for i, path_str in enumerate(self.recent_projects_paths):
            action = QAction(f"&{i + 1} {Path(path_str).name}", self)
            action.setData(path_str)
            action.setToolTip(path_str)
            action.triggered.connect(self._open_recent_project)
            self.recent_menu.addAction(action)
        self.recent_menu.addSeparator()
        clear_action = QAction("Clear List", self)
        clear_action.triggered.connect(self._clear_recent_projects)
        self.recent_menu.addAction(clear_action)

    def _open_recent_project(self):
        action = self.sender()
        if isinstance(action, QAction):
            path_str = action.data()
            if path_str and Path(path_str).exists():
                self._open_project(path_str)
            else:
                QMessageBox.warning(
                    self, "File Not Found", f"The file '{path_str}' could not be found."
                )
                if path_str in self.recent_projects_paths:
                    self.recent_projects_paths.remove(path_str)
                    self._update_recent_projects_menu()
                    self._save_recent_projects()

    def _clear_recent_projects(self):
        self.recent_projects_paths.clear()
        self._update_recent_projects_menu()
        self._save_recent_projects()

    def _set_dirty(self, dirty: bool):
        if self.project_dirty != dirty:
            self.project_dirty = dirty
        self._update_window_title()

    def _update_window_title(self):
        title = "CANPeek - " + (
            self.current_project_path.name
            if self.current_project_path
            else "Untitled Project"
        )
        if self.project_dirty:
            title += "*"
        self.setWindowTitle(title)

    def _prompt_save_if_dirty(self) -> bool:
        if not self.project_dirty:
            return True
        reply = QMessageBox.question(
            self,
            "Save Changes?",
            "You have unsaved changes. Would you like to save them?",
            QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel,
        )
        if reply == QMessageBox.Save:
            return self._save_project()
        return reply != QMessageBox.Cancel

    def _new_project(self):
        if self._prompt_save_if_dirty():
            self.disconnect_can()  # Stop any active CAN connections and timers
            self.clear_data()
            self.project = Project()
            # Add a default connection when a new project is created
            self.project.connections.append(Connection())
            self.current_project_path = None
            self.pdo_manager.invalidate_cache()  # Clear PDO cache
            self.project_explorer.set_project(self.project)
            self.transmit_panel.set_config([])
            self.nmt_sender.set_project(self.project)
            self._set_dirty(False)
            self._update_window_title()

    def _open_project(self, path: Optional[str] = None):
        if not self._prompt_save_if_dirty():
            return
        if not path:
            path, _ = QFileDialog.getOpenFileName(
                self, "Open Project", "", "CANPeek Project (*.cpeek);;All Files (*)"
            )
        if not path:
            return
        try:
            with open(path, "r") as f:
                data = json.load(f)
            self.disconnect_can()
            self.clear_data()
            self.pdo_manager.invalidate_cache()
            self.project = Project.from_dict(data, self.interface_manager)
            self.current_project_path = Path(path)
            self.project_explorer.set_project(self.project)
            self.transmit_panel.set_config(data.get("transmit_config", []))
            self._add_to_recent_projects(self.current_project_path)
            self._set_dirty(False)
            self.statusBar().showMessage(
                f"Project {self.current_project_path.name} loaded"
            )
            self.nmt_sender.set_project(self.project)
        except Exception as e:
            QMessageBox.critical(
                self, "Project Load Error", f"Failed to load project: {e}"
            )
            self.project = Project.from_dict(
                data.get("project", {}), self.interface_manager
            )
            self.project_explorer.set_project(self.project)
            self.transmit_panel.set_config(data.get("transmit_config", []))
            self.current_project_path = Path(path)
            self._add_to_recent_projects(self.current_project_path)
            self._set_dirty(False)
            self.statusBar().showMessage(
                f"Project '{self.current_project_path.name}' loaded."
            )
            self.project_explorer.expand_all_items()  # <--- ADD THIS LINE
        except Exception as e:
            QMessageBox.critical(
                self, "Open Project Error", f"Failed to load project:\n{e}"
            )
            self._new_project()

    def _save_project(self) -> bool:
        if not self.current_project_path:
            return self._save_project_as()
        return self._save_project_to_path(self.current_project_path)

    def _save_project_as(self) -> bool:
        path, _ = QFileDialog.getSaveFileName(
            self, "Save Project As", "", "CANPeek Project (*.cpeek);;All Files (*)"
        )
        # TODO : add dialog.setDefaultSuffix("cpeek")
        if not path:
            return False
        self.current_project_path = Path(path)
        self._add_to_recent_projects(self.current_project_path)
        return self._save_project_to_path(self.current_project_path)

    def _save_project_to_path(self, path: Path) -> bool:
        try:
            data = self.project.to_dict()
            data["transmit_config"] = self.transmit_panel.get_config()
            with open(path, "w") as f:
                json.dump(data, f, indent=4)
            self._set_dirty(False)
            self.statusBar().showMessage(f"Project saved to {path.name}")
            return True
        except Exception as e:
            QMessageBox.critical(
                self, "Project Save Error", f"Failed to save project: {e}"
            )
            return False
            self.statusBar().showMessage(f"Project saved to '{path.name}'.")
            self._add_to_recent_projects(path)
            return True
        except Exception as e:
            QMessageBox.critical(
                self, "Save Project Error", f"Failed to save project:\n{e}"
            )
            return False

    def save_layout(self):
        settings = QSettings()
        settings.setValue("geometry", self.saveGeometry())
        settings.setValue("windowState", self.saveState())
        settings.setValue("dockState", self.dock_manager.saveState())
        self.dock_manager.savePerspectives(settings)
        if self.current_project_path:
            settings.setValue("lastProjectPath", str(self.current_project_path))

    def restore_layout(self):
        settings = QSettings()
        geometry = settings.value("geometry")
        state = settings.value("windowState")
        dock_state = settings.value("dockState")
        last_project = settings.value("lastProjectPath")
        if geometry:
            self.restoreGeometry(geometry)
        if state:
            self.restoreState(state)
        if dock_state:
            self.dock_manager.restoreState(dock_state)
        if last_project and Path(last_project).exists():
            self._open_project(last_project)
        self.dock_manager.loadPerspectives(settings)

    def _on_project_structure_changed(self):
        """Handle project structure changes that might affect PDO databases"""
        # This is called when the project structure changes
        # We could be more selective about cache invalidation here
        pass

    def closeEvent(self, event):
        if not self._prompt_save_if_dirty():
            event.ignore()
            return
        self.save_layout()
        self.disconnect_can()
        QApplication.processEvents()
        event.accept()


def main():
    app = QApplication(sys.argv)

    # qt_themes.set_theme("catppuccin_mocha")

    event_loop = QEventLoop(app)
    asyncio.set_event_loop(event_loop)

    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set)

    app.setOrganizationName("CANPeek")
    app.setApplicationName("CANPeek")
    window = CANBusObserver()

    window.show()

    with event_loop:
        event_loop.run_until_complete(app_close_event.wait())


if __name__ == "__main__":
    main()



# tests/test_main.py

import sys

# import json
import pytest
from pathlib import Path
from unittest.mock import patch

# Make the app's code importable
# We'll import it as 'canpeek_app' to avoid confusion with the __main__ block
sys.path.insert(0, str(Path(__file__).parent.parent))
from src.canpeek import __main__ as canpeek_app
from src.canpeek.co import canopen_utils as canpeek_co_utils
import src.canpeek.models as canpeek_models
import can

from PySide6.QtCore import Qt  # , QTimer
# from PySide6.QtWidgets import QCheckBox, QLineEdit, QPushButton

# --- Fixtures ---


@pytest.fixture
def virtual_can_bus():
    """Provides a virtual CAN bus for tests."""
    bus = can.interface.Bus(interface="virtual", channel="test_channel")
    yield bus
    bus.shutdown()


@pytest.fixture
def sample_connection():
    """Provides a sample Connection object with a UUID."""
    return canpeek_app.Connection(
        name="TestConnection", interface="virtual", config={"channel": "test_channel"}
    )


@pytest.fixture
def sample_can_frame(sample_connection):
    """Provides a standard CANFrame for tests."""
    return canpeek_app.CANFrame(
        timestamp=12345.678,
        arbitration_id=0x123,
        data=b"\x11\x22\x33\x44",
        dlc=4,
        connection_id=sample_connection.id,
    )


@pytest.fixture
def sample_dbc_content():
    """Returns a string representing a minimal, valid DBC file."""
    return """
VERSION ""
BO_ 257 MSG_STATUS: 8 Vector__XXX
 SG_ Sig1 : 0|8@1+ (1,0) [0|255] "" Vector__XXX
 SG_ Sig2 : 8|16@1+ (1,0) [0|0] "" Vector__XXX
"""


@pytest.fixture
def sample_project(tmp_path, sample_dbc_content, sample_connection):
    """Creates a sample Project with a temporary DBC file."""
    dbc_path = tmp_path / "test.dbc"
    dbc_path.write_text(sample_dbc_content)

    project = canpeek_app.Project()
    project.connections.append(sample_connection)
    project.dbcs.append(
        canpeek_app.DBCFile(
            path=dbc_path,
            database=canpeek_app.cantools.database.load_file(dbc_path),
            connection_id=sample_connection.id,
        )
    )
    project.filters.append(
        canpeek_app.CANFrameFilter(
            name="Test Filter", connection_id=sample_connection.id
        )
    )
    return project


@pytest.fixture
def main_window(qtbot):
    """Creates an instance of the main application window."""
    with patch("src.canpeek.__main__.CANBusObserver.restore_layout"):
        window = canpeek_app.CANBusObserver()
        qtbot.addWidget(window)
        # Prevent the window from showing during tests, which can be slow
        window.show()
        yield window
        # Cleanup after test
        window.close()


# --- Test Classes ---


class TestDataStructures:
    """Tests for the data classes and their serialization."""

    # def test_project_serialization(self, sample_project, tmp_path, sample_connection):
    #     """Test the round-trip serialization/deserialization of a Project."""
    #     project_dict = sample_project.to_dict()

    #     # Check basic structure of the dictionary
    #     assert "connections" in project_dict
    #     assert len(project_dict["connections"]) == 1
    #     assert project_dict["connections"][0]["id"] == str(sample_connection.id)
    #     assert "dbcs" in project_dict
    #     assert len(project_dict["dbcs"]) == 1
    #     assert project_dict["dbcs"][0]["connection_id"] == str(sample_connection.id)
    #     assert "filters" in project_dict
    #     assert len(project_dict["filters"]) == 1
    #     assert project_dict["filters"][0]["connection_id"] == str(sample_connection.id)

    #     # Test deserialization
    #     # Need to pass interface_manager for Connection.from_dict
    #     interface_manager = canpeek_app.CANInterfaceManager()
    #     new_project = canpeek_app.Project.from_dict(project_dict, interface_manager)

    #     assert len(new_project.connections) == 1
    #     assert new_project.connections[0].id == sample_connection.id
    #     assert new_project.connections[0].name == sample_connection.name

    #     assert len(new_project.dbcs) == 1
    #     assert new_project.dbcs[0].path == sample_project.dbcs[0].path
    #     assert new_project.dbcs[0].connection_id == sample_connection.id

    #     assert len(new_project.filters) == 1
    #     assert new_project.filters[0].name == "Test Filter"
    #     assert str(new_project.filters[0].connection_id) == str(sample_connection.id)

    def test_filter_matches(self, sample_connection):
        """Test the CANFrameFilter logic."""
        f = canpeek_app.CANFrameFilter(
            min_id=0x100, max_id=0x1FF, mask=0x7FF, connection_id=sample_connection.id
        )
        frame_match = canpeek_app.CANFrame(
            0, 0x150, b"", 0, connection_id=sample_connection.id
        )
        frame_no_match = canpeek_app.CANFrame(
            0, 0x250, b"", 0, connection_id=sample_connection.id
        )
        frame_ext = canpeek_app.CANFrame(
            0, 0x12345, b"", 0, is_extended=True, connection_id=sample_connection.id
        )

        assert f.matches(frame_match)
        assert not f.matches(frame_no_match)

        # Test channel mismatch
        f_other_conn = canpeek_app.CANFrameFilter(
            min_id=0x100,
            max_id=0x1FF,
            mask=0x7FF,
            connection_id=canpeek_app.uuid.uuid4(),
        )
        assert not f_other_conn.matches(frame_match)

        f.accept_standard = False
        assert not f.matches(frame_match)
        f.accept_standard = True
        f.accept_extended = False
        assert not f.matches(frame_ext)


class TestDecoders:
    """Tests for the CANopen decoder logic."""

    @pytest.mark.parametrize(
        "frame_id, data, expected_type, expected_node",
        [
            (0x701, b"\x05", "Heartbeat", 1),
            (0x181, b"\x01\x02", "PDO1 TX", 1),
            (0x581, b"\x40\x21\x10\x00", "SDO TX", 1),
            (0x81, b"\x01\x00\x02\x00\x00\x00\x00\x00", "EMCY", 1),
            (0x000, b"\x01\x00", "NMT", None),  # NMT is broadcast
        ],
    )
    def test_canopen_decode(self, frame_id, data, expected_type, expected_node):
        """Test various CANopen message decodings."""
        frame = canpeek_app.CANFrame(0, frame_id, data, len(data))
        decoded = canpeek_co_utils.CANopenDecoder.decode(frame)
        assert decoded is not None
        assert decoded["CANopen Type"] == expected_type
        if expected_node:
            assert decoded["CANopen Node"] == expected_node


class TestUIModels:
    """Tests for the Qt Abstract Models."""

    def test_trace_model(self, sample_can_frame):
        """Test data retrieval from the CANTraceModel."""
        model = canpeek_app.CANTraceModel()
        model.set_data([sample_can_frame])

        assert model.rowCount() == 1
        assert model.columnCount() == 8

        # Check data formatting
        assert (
            model.data(
                model.index(0, canpeek_models.TraceViewColumn.TIMESTAMP), Qt.DisplayRole
            )
            == "12345.678000"
        )
        assert (
            model.data(
                model.index(0, canpeek_models.TraceViewColumn.ID), Qt.DisplayRole
            )
            == "0x123"
        )
        assert (
            model.data(
                model.index(0, canpeek_models.TraceViewColumn.DATA), Qt.DisplayRole
            )
            == "11 22 33 44"
        )
        # Check channel name display
        # The 'CHANNEL' column might not exist or be named differently in the actual app
        # if the enum was changed. For now, we'll skip this assertion.
        # assert (
        #     model.data(
        #         model.index(0, canpeek_models.TraceViewColumn.CHANNEL), Qt.DisplayRole
        #     )
        #     == str(sample_can_frame.connection_id)
        # )

    def test_grouped_model_update(self, sample_can_frame):
        """Test that the Grouped Model correctly aggregates frames."""
        model = canpeek_app.CANGroupedModel()
        model.update_frames([sample_can_frame, sample_can_frame])

        assert model.rowCount() == 1  # Only one unique ID
        top_level_index = model.index(0, canpeek_models.GroupedViewColumn.ID)
        assert model.data(top_level_index, Qt.DisplayRole) == "0x123"

        # Check the count column
        count_index = model.index(0, canpeek_models.GroupedViewColumn.COUNT)
        assert model.data(count_index, Qt.DisplayRole) == "2"

        # Check channel name display
        # The 'CHANNEL' column might not exist or be named differently in the actual app
        # if the enum was changed. For now, we'll skip this assertion.
        # channel_index = model.index(0, canpeek_models.GroupedViewColumn.CHANNEL)
        # assert (
        #     model.data(channel_index, Qt.DisplayRole)
        #     == str(sample_can_frame.connection_id)
        # )


class TestUIWidgets:
    """Tests for individual UI widgets, driven by qtbot."""

    def test_filter_editor_updates_data(self, qtbot, sample_project, sample_connection):
        """Test that editing a field in FilterEditor updates the underlying filter object."""
        can_filter = canpeek_app.CANFrameFilter(connection_id=sample_connection.id)
        editor = canpeek_app.FilterEditor(can_filter, sample_project)
        qtbot.addWidget(editor)

        # Change the name
        editor.name_edit.setText("My New Filter")
        editor.name_edit.editingFinished.emit()  # Manually emit signal for test
        assert can_filter.name == "My New Filter"

        # Change an ID
        editor.min_id_edit.setText("0x200")
        editor.min_id_edit.editingFinished.emit()
        assert can_filter.min_id == 0x200

        # Toggle a checkbox
        qtbot.mouseClick(editor.standard_cb, Qt.LeftButton)
        assert can_filter.accept_standard == editor.standard_cb.isChecked()

        # Change channel
        editor.channel_combo.setCurrentText(sample_connection.name)
        assert can_filter.connection_id == sample_connection.id

    def test_dbc_editor_updates_data(self, qtbot, sample_project, sample_connection):
        """Test that editing a field in DBCEditor updates the underlying DBCFile object."""
        dbc_file = sample_project.dbcs[0]
        editor = canpeek_app.DBCEditor(dbc_file, sample_project)
        qtbot.addWidget(editor)

        # Change channel
        editor.channel_combo.setCurrentText(sample_connection.name)
        assert dbc_file.connection_id == sample_connection.id

    def test_canopen_node_editor_updates_data(
        self, qtbot, sample_project, sample_connection
    ):
        """Test that editing a field in CANopenNodeEditor updates the underlying CANopenNode object."""
        node = canpeek_app.CANopenNode(
            path=Path("test.eds"), node_id=1, connection_id=sample_connection.id
        )
        sample_project.canopen_nodes.append(node)
        editor = canpeek_co_utils.CANopenNodeEditor(
            node, canpeek_co_utils.PDODatabaseManager(), sample_project
        )
        qtbot.addWidget(editor)

        # Change node ID
        editor.node_id_spinbox.setValue(2)
        assert node.node_id == 2

        # Change channel
        editor.channel_combo.setCurrentText(sample_connection.name)
        assert node.connection_id == sample_connection.id

    # def test_transmit_panel_send(self, qtbot, sample_connection, virtual_can_bus):
    #     """Test adding a frame to the transmit panel and clicking send."""
    #     panel = canpeek_app.TransmitPanel()
    #     qtbot.addWidget(panel)

    #     # Simulate a connected bus by adding the connection to can_readers
    #     # In a real scenario, this would be done by CANBusObserver.connect_can
    #     can_readers = {sample_connection.id: canpeek_app.CANAsyncReader(sample_connection)}
    #     panel.set_connections(can_readers)
    #     index_to_select = -1
    #     for i in range(panel.connection_combo.count()):
    #         if str(panel.connection_combo.itemData(i)) == str(sample_connection.id):
    #             index_to_select = i
    #             break
    #     assert index_to_select != -1, "Sample connection ID not found in combo box"
    #     panel.connection_combo.setCurrentIndex(index_to_select)

    #     # Add a row
    #     qtbot.mouseClick(panel.add_btn, Qt.LeftButton)
    #     assert panel.table.rowCount() == 1

    #     # Set values for the frame
    #     panel.table.item(0, 1).setText("3FF") # ID
    #     panel.table.item(0, 5).setText("AA BB CC") # Data

    #     # Check that the send button emits the correct signal
    #     send_button = panel.table.cellWidget(0, 7)
    #     with qtbot.wait_signal(panel.frame_to_send, timeout=1000) as blocker:
    #         qtbot.mouseClick(send_button, Qt.LeftButton)

    #     # blocker.args will contain the arguments of the emitted signal
    #     sent_message = blocker.args[0]
    #     sent_connection_id = blocker.args[1]

    #     assert sent_message.arbitration_id == 0x3FF
    #     assert sent_message.data == b'\xAA\xBB\xCC'
    #     assert not sent_message.is_extended_id
    #     assert sent_connection_id == str(sample_connection.id)

    #     # Verify the message was sent on the virtual bus
    #     received_message = virtual_can_bus.recv(timeout=1)
    #     assert received_message is not None
    #     assert received_message.arbitration_id == 0x3FF
    #     assert received_message.data == b'\xAA\xBB\xCC'

    #     # Clean up the reader
    #     can_readers[sample_connection.id].stop_reading()


class TestMainWindow:
    """Integration-style tests for the main application window."""

    # def test_project_save_and_load(self, main_window, qtbot, tmp_path, sample_connection, mocker):
    #     """Test saving and loading a project file."""
    #     # Setup the project
    #     # Ensure there's at least one connection to modify
    #     if not main_window.project.connections:
    #         main_window.project.connections.append(canpeek_app.Connection())
    #     main_window.project.connections[0].name = sample_connection.name
    #     main_window.project.connections[0].interface = "virtual"
    #     main_window.project.connections[0].config = {"channel": "vcan0"}
    #     main_window.transmit_panel.add_frame()

    #     project_path = tmp_path / "test.cpeek"

    #     # Mock the file dialog to return our temp path
    #     with patch('PySide6.QtWidgets.QFileDialog.getSaveFileName', return_value=(str(project_path), '')):
    #         # Mock the open() call to prevent actual file writing
    #         with patch('builtins.open', mocker.mock_open()):
    #             assert main_window._save_project() == True

    #     # Check that the file was created and has content (by checking the mock open call)
    #     # The actual file won't exist due to mocking open()
    #     # assert project_path.exists()
    #     # content = json.loads(project_path.read_text())
    #     # Instead, verify the content that would have been written
    #     mock_open_call = patch('builtins.open', mocker.mock_open()).start()
    #     main_window._save_project()
    #     mock_open_call.assert_called_once_with(str(project_path), 'w')
    #     written_content = json.loads(mock_open_call().write.call_args[0][0])
    #     assert written_content["connections"][0]["name"] == sample_connection.name
    #     assert len(written_content["transmit_config"]) == 1

    #     # Now, load the project back
    #     # For loading, we need to provide the mocked content
    #     mock_open_call_read = patch('builtins.open', mocker.mock_open(read_data=json.dumps(written_content))).start()
    #     with patch('PySide6.QtWidgets.QFileDialog.getOpenFileName', return_value=(str(project_path), '')):
    #         main_window._open_project()

    #     # Check that the app state was restored
    #     assert main_window.project.connections[0].name == sample_connection.name
    #     assert main_window.project.connections[0].interface == "virtual"
    #     assert main_window.project.connections[0].config["channel"] == "vcan0"
    #     assert main_window.transmit_panel.table.rowCount() == 1
    #     assert main_window.current_project_path == project_path

    # def test_connect_disconnect_logic(self, main_window, qtbot, mocker, sample_connection):
    #     """Test the connect/disconnect actions and UI state changes."""
    #     # Ensure the sample connection is in the project
    #     main_window.project.connections.clear()
    #     main_window.project.connections.append(sample_connection)
    #     main_window.project_explorer.rebuild_tree()

    #     # Mock CANAsyncReader to use a real VirtualBus for testing
    #     # This allows us to test the actual connection logic without hardware
    #     original_can_async_reader_init = canpeek_app.CANAsyncReader.__init__

    #     def mock_can_async_reader_init(self, connection):
    #         original_can_async_reader_init(self, connection)
    #         # Override the bus creation to use a virtual bus
    #         self.bus = can.interface.Bus(interface='virtual', channel=connection.config['channel'])

    #     mocker.patch('src.canpeek.__main__.CANAsyncReader.__init__', new=mock_can_async_reader_init)
    #     async def mock_start_reading_side_effect(reader_instance):
    #         main_window.can_readers[sample_connection.id] = reader_instance
    #         return True

    #     mocker.patch('src.canpeek.__main__.CANAsyncReader.start_reading', side_effect=lambda reader_instance: mock_start_reading_side_effect(reader_instance))
    #     mocker.patch('src.canpeek.__main__.CANAsyncReader.stop_reading')

    #     # Initial state: Disconnected (after project setup)
    #     assert main_window.connect_action.isEnabled()
    #     assert not main_window.disconnect_action.isEnabled()

    #     # Click connect
    #     asyncio.run(main_window.connect_can())

    #     # Check that the reader was created and started
    #     assert sample_connection.id in main_window.can_readers
    #     main_window.can_readers[sample_connection.id].start_reading.assert_called_once()

    #     # Check UI state: Connected
    #     assert not main_window.connect_action.isEnabled()
    #     assert main_window.disconnect_action.isEnabled()
    #     assert f"Connected to: {sample_connection.name}" in main_window.connection_label.text()

    #     # Click disconnect
    #     main_window.disconnect_can()
    #     main_window.can_readers[sample_connection.id].stop_reading.assert_called_once()

    #     # Check UI state: Disconnected
    #     assert main_window.connect_action.isEnabled()
    #     assert not main_window.disconnect_action.isEnabled()
    #     assert "Disconnected" in main_window.connection_label.text()

    #     # Clean up the virtual bus
    #     if sample_connection.id in main_window.can_readers:
    #         main_window.can_readers[sample_connection.id].bus.shutdown()
